import sys
from typing import Any, List, Tuple

from astroid import nodes
from pytest import skip

from python_ta.typecheck.base import TypeFailAnnotationUnify

from .. import custom_hypothesis_support as cs
from ..custom_hypothesis_support import lookup_type


def test_single_annotation_int():
    src = """
    def foo(x: int):
        return x
    """
    ast_mod, ti = cs._parse_text(src)

    func_node = next(ast_mod.nodes_of_class(nodes.FunctionDef))
    assert lookup_type(ti, func_node, "x") == int


def test_single_annotation_str():
    src = """
    def foo(x: str):
        return x
    """
    ast_mod, ti = cs._parse_text(src)

    func_node = next(ast_mod.nodes_of_class(nodes.FunctionDef))
    assert lookup_type(ti, func_node, "x") == str


def test_multiple_annotations():
    src = """
    def foo(x: int, y: int):
        return x + y
    """
    ast_mod, ti = cs._parse_text(src)

    func_node = next(ast_mod.nodes_of_class(nodes.FunctionDef))
    assert lookup_type(ti, func_node, "x") == int
    assert lookup_type(ti, func_node, "y") == int


def test_multiple_annotations_diff_type():
    src = """
    def foo(x: int, y: str):
        print(y)
        return x
    """
    ast_mod, ti = cs._parse_text(src)

    func_node = next(ast_mod.nodes_of_class(nodes.FunctionDef))
    assert lookup_type(ti, func_node, "x") == int
    assert lookup_type(ti, func_node, "y") == str


def test_call_wrong_type():
    src = """
    def foo(x: int):
        return x

    foo('Hello')
    """
    ast_mod, ti = cs._parse_text(src)

    func_node = next(ast_mod.nodes_of_class(nodes.FunctionDef))
    assert lookup_type(ti, func_node, "x") == int

    call_node = next(ast_mod.nodes_of_class(nodes.Call))
    assert isinstance(call_node.inf_type, TypeFailAnnotationUnify)


def test_call_wrong_type_str():
    src = """
    def foo(x: str):
        return x

    foo(5)
    """
    ast_mod, ti = cs._parse_text(src)

    func_node = next(ast_mod.nodes_of_class(nodes.FunctionDef))
    assert lookup_type(ti, func_node, "x") == str

    call_node = next(ast_mod.nodes_of_class(nodes.Call))
    assert isinstance(call_node.inf_type, TypeFailAnnotationUnify)


def test_call_multiple_annotation_wrong_type():
    src = """
    def foo(x: int, y: str):
        return x

    foo('Hello', 'Goodbye')
    """
    ast_mod, ti = cs._parse_text(src)

    func_node = next(ast_mod.nodes_of_class(nodes.FunctionDef))
    assert lookup_type(ti, func_node, "x") == int

    call_node = next(ast_mod.nodes_of_class(nodes.Call))
    assert isinstance(call_node.inf_type, TypeFailAnnotationUnify)


def test_mixed_annotation():
    src = """
    def foo(x: int, y):
        return y

    foo(5, 'Hello')
    """
    ast_mod, ti = cs._parse_text(src)

    func_node = next(ast_mod.nodes_of_class(nodes.FunctionDef))
    assert lookup_type(ti, func_node, "x") == int
    assert lookup_type(ti, func_node, "y") == Any

    call_node = next(ast_mod.nodes_of_class(nodes.Call))
    assert call_node.inf_type.getValue() == Any


def test_mixed_annotation_wrong():
    src = """
    def foo(x: int, y):
        return y

    foo('Hello', 5)
    """
    ast_mod, ti = cs._parse_text(src)

    func_node = next(ast_mod.nodes_of_class(nodes.FunctionDef))
    assert lookup_type(ti, func_node, "x") == int
    assert lookup_type(ti, func_node, "y") == Any

    call_node = next(ast_mod.nodes_of_class(nodes.Call))
    assert isinstance(call_node.inf_type, TypeFailAnnotationUnify)


def test_param_subscript_list():
    src = """
    def foo(lst: List):
        return lst

    foo([1, 2, 3])
    """
    ast_mod, ti = cs._parse_text(src)

    func_node = next(ast_mod.nodes_of_class(nodes.FunctionDef))
    assert lookup_type(ti, func_node, "lst") == List[Any]

    call_node = next(ast_mod.nodes_of_class(nodes.Call))
    assert call_node.inf_type.getValue() == List[Any]


def test_param_subscript_list_int():
    src = """
    def foo(lst: List[int]):
        return lst

    foo([1, 2, 3])
    """
    ast_mod, ti = cs._parse_text(src)

    func_node = next(ast_mod.nodes_of_class(nodes.FunctionDef))
    assert lookup_type(ti, func_node, "lst") == List[int]

    call_node = next(ast_mod.nodes_of_class(nodes.Call))
    assert call_node.inf_type.getValue() == List[int]


def test_param_subscript_tuple():
    src = """
    def foo(t: Tuple[int, int]):
        return t

    foo((0, 1))
    """
    skip("Requires support for multi-parameter Tuple annotations")
    ast_mod, ti = cs._parse_text(src)

    func_node = next(ast_mod.nodes_of_class(nodes.FunctionDef))
    assert lookup_type(ti, func_node, "t") == Tuple[int, int]

    call_node = next(ast_mod.nodes_of_class(nodes.Call))
    assert call_node.inf_type.getValue() == Tuple[int, int]


def test_return_list():
    src = """
    def foo(x) -> List:
        return [x]

    foo(0)
    """
    ast_mod, ti = cs._parse_text(src)

    call_node = next(ast_mod.nodes_of_class(nodes.Call))
    assert call_node.inf_type.getValue() == List[Any]


def test_return_list_subscript():
    src = """
    def foo(x: int) -> List[int]:
        return [x]

    foo(0)
    """
    ast_mod, ti = cs._parse_text(src)

    call_node = next(ast_mod.nodes_of_class(nodes.Call))
    assert call_node.inf_type.getValue() == List[int]



from astroid import nodes
from hypothesis import settings
from pytest import skip

from python_ta.typecheck.base import TypeFail

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


def test_classdef_attribute_assign():
    """Test whether type of attributes are properly being set."""
    program = (
        f"class Network:\n"
        f"    def __init__(self, name, id):\n"
        f"        self.name = name\n"
        f"        self.id = id"
        f"\n"
        f'rogers = Network("Rogers", 5)\n'
        f'rogers.name = "BoB"\n'
        f'self = Network("asdf", 5)\n'
        f'self.name = "asdfaBoB"\n'
        f"\n"
    )
    module, inferer = cs._parse_text(program)
    classdef_node = next(module.nodes_of_class(nodes.ClassDef))
    for attribute_lst in classdef_node.instance_attrs.values():
        for instance in attribute_lst:
            attribute_type = inferer.type_constraints.resolve(
                classdef_node.type_environment.lookup_in_env(instance.attrname)
            )
            value_type = inferer.type_constraints.resolve(instance.parent.value.inf_type.getValue())
            assert attribute_type == value_type


def test_classdef_method_call():
    """Test whether type of the method call are properly being set"""
    program = (
        f"class Network:\n"
        f"    def __init__(self, name):\n"
        f"        self.name = name\n"
        f"    def get_name(self):\n"
        f"        return self.name\n "
        f"\n"
        f'rogers = Network("Rogers")\n'
        f"rogers.get_name()"
        f"\n"
    )
    module, inferer = cs._parse_text(program, True)
    attribute_node = list(module.nodes_of_class(nodes.Attribute))[1]
    expected_rtype = attribute_node.parent.inf_type.getValue()
    actual_rtype = inferer.type_constraints.resolve(
        attribute_node.inf_type.getValue().__args__[-1]
    ).getValue()
    assert actual_rtype.__name__ == expected_rtype.__name__


def test_classdef_method_call_annotated_concrete():
    """Test whether types of the method calls are properly being set given the annotations."""
    program = (
        f"class Network:\n"
        f"    def __init__(self, name: str) -> None:\n"
        f"        self.name = name\n"
        f"        status = 0\n"
        f"    def set_status(self, status: int) -> int:\n"
        f"        self.status = status\n"
        f"        return self.status\n"
        f"\n"
    )
    module, inferer = cs._parse_text(program)
    for functiondef_node in module.nodes_of_class(nodes.FunctionDef):
        self_name = functiondef_node.args.args[0].name
        actual_type = inferer.type_constraints.resolve(
            functiondef_node.type_environment.lookup_in_env(self_name)
        ).getValue()
        assert actual_type.__forward_arg__ == functiondef_node.parent.name
        for i in range(1, len(functiondef_node.args.annotations)):
            arg_name = functiondef_node.args.args[i].name
            actual_type = inferer.type_constraints.resolve(
                functiondef_node.type_environment.lookup_in_env(arg_name)
            ).getValue()
            assert actual_type.__name__ == functiondef_node.args.annotations[i].name
        expected_rtype = inferer.type_constraints.resolve(
            functiondef_node.parent.type_environment.lookup_in_env(functiondef_node.name)
        ).getValue()
        if functiondef_node.name == "__init__":
            class_type = inferer.type_constraints.resolve(
                functiondef_node.type_environment.lookup_in_env(self_name)
            ).getValue()
            assert class_type == expected_rtype.__args__[-1]
        else:
            assert functiondef_node.returns.name == expected_rtype.__args__[-1].__name__


def test_bad_attribute_access():
    """User tries to access a non-existing attribute; or misspells the attribute name."""
    program = f"x = 1\n" f"x.wrong_name\n"
    module, inferer = cs._parse_text(program)
    expr_node = next(module.nodes_of_class(nodes.Expr))
    expected_msg = "TypeFail: Invalid attribute lookup x.wrong_name"
    assert expr_node.inf_type.getValue() == expected_msg


def test_builtin_method_call_bad_self():
    """User tries to call a method on an object of the wrong type (self)."""
    program = f"x = 1\n" f"x.append(1.0)\n"
    module, inferer = cs._parse_text(program)
    call_node = next(module.nodes_of_class(nodes.Call))
    expected_msg = f"TypeFail: Invalid attribute lookup x.append(1.0)"
    assert isinstance(call_node.inf_type, TypeFail)
    assert call_node.inf_type.getValue() == expected_msg


def test_builtin_method_call_bad_argument():
    """User tries to call a method on an argument of the wrong type."""
    program = f"x = [1, 2, 3]\n" f"x.extend(1)\n"
    module, inferer = cs._parse_text(program)
    call_node = next(module.nodes_of_class(nodes.Call))
    assert isinstance(call_node.inf_type, TypeFail)



from typing import Any

from astroid import nodes
from hypothesis import HealthCheck, given, settings

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


@given(cs.ifexp_node())
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_ifexp(node):
    """Test the type setting of an IfExp node representing an if statement."""
    module, type_inferer = cs._parse_text(node)
    for ifexp_node in module.nodes_of_class(nodes.IfExp):
        assert ifexp_node.inf_type.getValue() == ifexp_node.body.inf_type.getValue()



import sys

from colorama import Back, Fore, Style, colorama_text
from pylint.interfaces import IReporter

from .plain_reporter import PlainReporter


class ColorReporter(PlainReporter):
    """Colorized text reporter. Should only be used to print to stdout."""

    __implements__ = IReporter
    name = "ColorReporter"

    _COLOURING = {
        "black": Fore.BLACK,
        "black-line": Fore.BLACK,
        "bold": Style.BRIGHT,
        "code-heading": Fore.RED + Style.BRIGHT,
        "style-heading": Fore.BLUE + Style.BRIGHT,
        "code-name": Fore.RED + Style.BRIGHT,
        "style-name": Fore.BLUE + Style.BRIGHT,
        "highlight": Style.BRIGHT + Fore.BLACK + Back.CYAN,
        "grey": Fore.LIGHTBLACK_EX,
        "grey-line": Fore.LIGHTBLACK_EX,
        "gbold": Style.BRIGHT + Fore.LIGHTBLACK_EX,
        "gbold-line": Style.BRIGHT + Fore.LIGHTBLACK_EX,
        "reset": Style.RESET_ALL,
    }

    def print_messages(self, level: str = "all") -> None:
        """Print messages for the current file.

        If level == 'all', both errors and style errors are displayed. Otherwise,
        only errors are displayed.
        """
        # Check if the OS currently running is Windows
        with colorama_text(wrap=(sys.platform == "win32"), strip=False):
            super().print_messages(level)

    @classmethod
    def _colourify(cls, colour_class: str, text: str) -> str:
        """
        Adds given ANSI colouring tokens (or key to colouring tokens in the
        class-level dict "_COLOURING") to text as well as final colour reset.

        Does not colour indents, except non-space indents.
        Called by _colour_messages_by_type and _add_line.
        """
        colour = cls._COLOURING[colour_class]
        new_text = text.lstrip(" ")
        space_count = len(text) - len(new_text)
        new_text = new_text.replace(" ", cls._SPACE)
        return (space_count * cls._SPACE) + colour + new_text + cls._COLOURING["reset"]



"""checker for use of I/O functions.
"""

from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker

FORBIDDEN_BUILTIN = ["input", "print", "open"]


class IOFunctionChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = "IO_Function"
    msgs = {
        "E9998": (
            "Used input/output function %s",
            "forbidden-IO-function",
            'Used when you use the I/O functions "print", "open" or "input". These '
            "functions should not be used except where specified by your instructor.",
        )
    }
    options = (
        (
            "forbidden-io-functions",
            {
                "default": FORBIDDEN_BUILTIN,
                "type": "csv",
                "metavar": "<builtin function names>",
                "help": "List of built-in function names that should not be used.",
            },
        ),
        (
            "allowed-io",
            {
                "default": (),
                "type": "csv",
                "metavar": "<forbidden io>",
                "help": "Functions where an I/O function may be used.",
            },
        ),
    )

    # this is important so that your checker is executed before others
    priority = -1

    @check_messages("forbidden-IO-function")
    def visit_call(self, node):
        if isinstance(node.func, nodes.Name):
            name = node.func.name
            # ignore the name if it's not a builtin (i.e. not defined in the
            # locals nor globals scope)
            if not (name in node.frame() or name in node.root()):
                scope = node.scope()
                # TODO: Only FunctionDefs are checked. Include global scope?
                if (
                    isinstance(scope, nodes.FunctionDef)
                    and scope.name not in self.config.allowed_io
                ):
                    if name in self.config.forbidden_io_functions:
                        self.add_message("forbidden-IO-function", node=node, args=name)


def register(linter):
    linter.register_checker(IOFunctionChecker(linter))



from typing import List

from astroid import nodes
from hypothesis import HealthCheck, assume, given, settings

from python_ta.typecheck.base import TypeFail, TypeFailFunction
from python_ta.typecheck.errors import BINOP_TO_METHOD

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


@given(cs.binop_node())
@settings(suppress_health_check=[HealthCheck.too_slow, HealthCheck.filter_too_much])
def test_binop_non_bool_concrete(node):
    """Test type setting of BinOp node(s) with non-boolean operands."""
    module, inferer = cs._parse_text(node)
    binop_node = list(module.nodes_of_class(nodes.BinOp))[0]
    left_type, right_type = (
        binop_node.left.inf_type.getValue(),
        binop_node.right.inf_type.getValue(),
    )
    exp_func_type = inferer.type_store.lookup_method(
        BINOP_TO_METHOD[node.op], left_type, right_type, node=binop_node
    )
    if not isinstance(exp_func_type, TypeFailFunction):
        exp_return_type = exp_func_type.__args__[-1]
    else:
        exp_return_type = None
    assume(exp_return_type is not None)
    assert binop_node.inf_type.getValue() == exp_return_type


def test_binop_reverse():
    src = """
    x = 3 * [1,2,3]
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    x = [ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)][0]
    assert ti.type_constraints.resolve(x).getValue() == List[int]


def test_binop_autoconvert():
    program = """
    x = 1 + 1.0
    y = 1 + 1j
    z = 1.0 + 1j
    """
    module, inferer = cs._parse_text(program, reset=True)
    x, y, z = [
        inferer.lookup_typevar(node, node.name) for node in module.nodes_of_class(nodes.AssignName)
    ]
    assert inferer.type_constraints.resolve(x).getValue() == float
    assert inferer.type_constraints.resolve(y).getValue() == complex
    assert inferer.type_constraints.resolve(z).getValue() == complex


def test_binop_autoconvert_fail():
    program = """
    x = 1 << 1.0
    y = 1.0 << 1
    z = 1j % 2j
    w = 1j << 2
    """
    module, inferer = cs._parse_text(program, reset=True)
    for binop_node in module.nodes_of_class(nodes.BinOp):
        assert isinstance(binop_node.inf_type, TypeFail)


def test_binop_userdefn():
    src = """
    class A:
        def __add__(self, i):
            return 1

        def __radd__(self, i):
            return 1.0

    class B:
        def __add__(self, i):
            return 'abc'

        def __radd__(self, i):
            return True

    a = A()
    b = B()
    x = a + b
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    a, b, x = [
        ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)
    ][8:]
    assert ti.type_constraints.resolve(x).getValue() == int


def test_binop_reverse_userdefn():
    src = """
    class A:
        def __radd__(self, i):
            return 1.0

    class B:
        def __add__(self, i):
            return 'abc'

        def __radd__(self, i):
            return True

    a = A()
    b = B()
    x = a + b
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    a, b, x = [
        ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)
    ][6:]
    assert ti.type_constraints.resolve(x).getValue() == bool


def test_binop_reverse_right_subclasses_left():
    src = """
    class A:
        def __add__(self, i):
            return 1

        def __radd__(self, i):
            return 1.0

    class B(A):
        def __add__(self, i):
            return 'abc'

        def __radd__(self, i):
            return [1]

    a = A()
    b = B()
    x = a + b
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    a, b, x = [
        ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)
    ][8:]
    assert ti.type_constraints.resolve(x).getValue() == List[int]



from astroid import nodes
from hypothesis import HealthCheck, assume, given, settings

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


@given(cs.unaryop_node(cs.non_bool_unary_op, cs.const_node(cs.numeric_values)))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_unaryop_non_bool_concrete(node):
    """Test type setting of UnaryOp node(s) with non-boolean operand."""
    assume(not (node.op == "~" and isinstance(node.operand.value, float)))
    module, _ = cs._parse_text(node)
    unaryop_node = list(module.nodes_of_class(nodes.UnaryOp))[0]
    assert unaryop_node.inf_type.getValue() == unaryop_node.operand.inf_type.getValue()


@given(cs.unaryop_node(cs.unary_bool_operator))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_not_bool(node):
    """Test type setting of UnaryOp node representing boolean operation not."""
    module, _ = cs._parse_text(node)
    for unaryop_node in module.nodes_of_class(nodes.UnaryOp):
        if unaryop_node.op == "not":
            assert unaryop_node.inf_type.getValue() == bool



"""Patch to add transforms for setting type constraints and creating control flow graphs.
"""
from pylint.lint import PyLinter

# from ..transforms.type_inference_visitor import TypeInferer
from ..cfg.visitor import CFGVisitor


def patch_ast_transforms():
    old_get_ast = PyLinter.get_ast

    def new_get_ast(self, filepath, modname):
        ast = old_get_ast(self, filepath, modname)
        if ast is not None:
            # type_inferer = TypeInferer()
            # env_transformer = type_inferer.environment_transformer()
            # type_transformer = type_inferer.type_inference_transformer()
            try:
                ast.accept(CFGVisitor())
                # env_transformer.visit(ast)
                # type_transformer.visit(ast)
            except:
                pass
        return ast

    PyLinter.get_ast = new_get_ast



"""Patch pylint message-handling behaviour."""
from astroid.transforms import TransformVisitor
from pylint.interfaces import UNDEFINED
from pylint.lint import PyLinter
from pylint.message import MessagesHandlerMixIn

from python_ta.transforms.setendings import register_transforms


def patch_messages():
    """Patch MessagesHandlerMixIn to pass the node to reporter."""
    old_add_message = MessagesHandlerMixIn.add_message

    def new_add_message(
        self, msg_id, line=None, node=None, args=None, confidence=UNDEFINED, col_offset=None
    ):
        old_add_message(self, msg_id, line, node, args, confidence, col_offset)
        msg_info = self.msgs_store.get_message_definitions(msg_id)[0]
        self.reporter.handle_node(msg_info, node)

    MessagesHandlerMixIn.add_message = new_add_message


def patch_linter_transform():
    """Patch PyLinter class to apply message transform with source code."""
    old_get_ast = PyLinter.get_ast

    def new_get_ast(self, filepath, modname):
        ast = old_get_ast(self, filepath, modname)
        if ast is not None:
            with open(filepath, encoding="utf-8") as f:
                source_code = f.readlines()
            ending_transformer = TransformVisitor()
            register_transforms(source_code, ending_transformer)
            ending_transformer.visit(ast)
        return ast

    PyLinter.get_ast = new_get_ast



from typing import Any, Callable, Tuple

from astroid import nodes
from hypothesis import HealthCheck, assume, given, settings
from pytest import skip

from python_ta.typecheck.base import NoType

from .. import custom_hypothesis_support as cs
from ..custom_hypothesis_support import lookup_type

settings.load_profile("pyta")


@given(cs.homogeneous_list(min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_for_homogeneous_list(iterable):
    """Test whether visitors properly set the type constraint of the a For node representing for/else statement
    iterating over a homogeneous list.
    """
    program = f"for elt in {iterable}:\n" f"    x = elt\n"
    module, TypeInferrer = cs._parse_text(program)
    for_node = list(module.nodes_of_class(nodes.For))[0]
    local_type_var = module.type_environment.lookup_in_env("x")
    inferred_type = TypeInferrer.type_constraints.resolve(local_type_var).getValue()
    assert inferred_type == for_node.iter.inf_type.getValue().__args__[0]


@given(cs.random_list(min_size=2))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_for_heterogeneous_list(iterable):
    """Test whether visitors properly set the type constraint of the a For node representing for/else statement
    iterating over a heterogeneous list.
    """
    assume(type(iterable[0]) != type(iterable[1]))
    val_types = [type(val) for val in iterable]
    if int in val_types:
        assume(bool not in val_types)
    if bool in val_types:
        assume(int not in val_types)
    program = f"for elt in {iterable}:\n" f"    x = elt\n"
    module, TypeInferrer = cs._parse_text(program)
    for_node = list(module.nodes_of_class(nodes.For))[0]
    local_type_var = module.type_environment.lookup_in_env("x")
    inferred_type = TypeInferrer.type_constraints.resolve(local_type_var).getValue()
    assert inferred_type == Any


def test_inference_func_def_for():
    """Test whether visitors properly set the type constraint of the a For node representing for/else statement
    iterating over a more complex iterable (ie, tuples, dicts, nested iterables).
    """
    program = (
        f"def add_ten(x):\n"
        f"    for num in [1, 2, 3, 4]:\n"
        f"        x = x + num\n"
        f"    return x\n"
    )
    module, TypeInferrer = cs._parse_text(program)
    for_node = list(module.nodes_of_class(nodes.For))[0]
    function_def_node = list(module.nodes_of_class(nodes.FunctionDef))[0]
    function_type_var = module.type_environment.lookup_in_env(function_def_node.name)
    function_type = TypeInferrer.type_constraints.resolve(function_type_var).getValue()
    target_type_var = function_def_node.type_environment.lookup_in_env("num")
    target_type = TypeInferrer.type_constraints.resolve(target_type_var).getValue()
    assert (function_type == Callable[[int], int]) and (target_type == int)


def test_for_list_tuple():
    program = """
        some_list = [('A', 1), ('B', 2)]

        for elt in some_list:
            x = elt
        """
    module, ti = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.AssignName):
        if assign_node.name == "x" or assign_node.name == "elt":
            assert lookup_type(ti, assign_node, assign_node.name) == Tuple[str, int]


def test_for_list_tuple_multi_arg():
    program = """
        some_list = [('A', 1), ('B', 2)]

        for a, b in some_list:
            x = a
            y = b
        """
    module, ti = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.AssignName):
        if assign_node.name == "x" or assign_node.name == "a":
            assert lookup_type(ti, assign_node, assign_node.name) == str
        elif assign_node.name == "y" or assign_node.name == "b":
            assert lookup_type(ti, assign_node, assign_node.name) == int


def test_for_zip():
    program = """
        some_str_list = ['A', 'B']
        some_int_list = [1, 2]

        for a, b in zip(some_str_list, some_int_list):
            x = a
            y = b
        """
    module, ti = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.AssignName):
        if assign_node.name == "x" or assign_node.name == "a":
            assert lookup_type(ti, assign_node, assign_node.name) == str
        elif assign_node.name == "y" or assign_node.name == "b":
            assert lookup_type(ti, assign_node, assign_node.name) == int


def test_for_dict():
    program = """
        some_dict = {'A': 1, 'B': 2}

        for a, b in some_dict.items():
            x = a
            y = b
        """
    skip(
        f"Return type of some_dict.items() is inferred as ItemsView[str, int],"
        f"which does not unify with List[Tuple[str, int]]"
    )
    module, ti = cs._parse_text(program)
    for assign_node in module.nodes_of_class(nodes.AssignName):
        if assign_node.name == "x" or assign_node.name == "a":
            assert lookup_type(ti, assign_node, assign_node.name) == str
        elif assign_node.name == "y" or assign_node.name == "b":
            assert lookup_type(ti, assign_node, assign_node.name) == int


def test_for_target_attr():
    program = """
    class A:
        def __init__(self):
            self.attr = 0

    a = A()

    for a.attr in range(5):
        a.attr
    """
    module, ti = cs._parse_text(program)
    for_node = next(module.nodes_of_class(nodes.For))
    assert isinstance(for_node.inf_type, NoType)


def test_for_target_subscript():
    program = """
    lst = [0, 1, 2]

    for lst[0] in lst:
        lst[0]
    """
    module, ti = cs._parse_text(program)
    for_node = next(module.nodes_of_class(nodes.For))
    assert isinstance(for_node.inf_type, NoType)



from astroid import nodes
from pytest import skip

from python_ta.typecheck.base import TypeFail, TypeInfo

from .. import custom_hypothesis_support as cs


def test_lambda_simple_call():
    src = """
    y = (lambda x: x + 1)(0)
    z = (lambda x: x + 'def')('abc')
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for lambda_node in ast_mod.nodes_of_class(nodes.Lambda):
        assert (
            str(lambda_node.inf_type.getValue()) == "typing.Callable[[int], int]"
            or str(lambda_node.inf_type.getValue()) == "typing.Callable[[str], str]"
        )
    for var_node in ast_mod.nodes_of_class(nodes.AssignName):
        if var_node.name == "y":
            y = ti.lookup_typevar(var_node, "y")
            assert ti.type_constraints.resolve(y).getValue() == int
        if var_node.name == "z":
            y = ti.lookup_typevar(var_node, "z")
            assert ti.type_constraints.resolve(y).getValue() == str


def test_lambda_simple_call_invalid():
    src = """
    y = (lambda x: x + 1)('abc')
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert isinstance(call_node.inf_type, TypeFail)


def test_define_fn_lambda():
    src = """
    f = lambda x: x + 1
    y = f(0)
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for var_node in ast_mod.nodes_of_class(nodes.AssignName):
        if var_node.name == "y":
            y = ti.lookup_typevar(var_node, "y")
            assert ti.type_constraints.resolve(y).getValue() == int


def test_define_fn_lambda_invalid():
    src = """
    f = lambda x: x + 1
    x = f('abc')
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert isinstance(call_node.inf_type, TypeFail)


def test_lambda_polymorphic_simple():
    src = """
    f = lambda x: x
    y = f('abc')
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for lambda_node in ast_mod.nodes_of_class(nodes.Lambda):
        x = ti.lookup_inf_type(lambda_node, "x").getValue()
        assert str(lambda_node.inf_type.getValue()) == f"typing.Callable[[{x}], {x}]"
    for var_node in ast_mod.nodes_of_class(nodes.AssignName):
        if var_node.name == "y":
            y = ti.lookup_typevar(var_node, "y")
            assert ti.type_constraints.resolve(y).getValue() == str


def test_lambda_polymorphic_builtin_lookup():
    skip(
        "This requires proper handling of unresolved calls to"
        "builtin methods with multiple variants"
    )
    src = """
    f = lambda x, y: x + y
    y = f('abc', 'def')
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for lambda_node in ast_mod.nodes_of_class(nodes.Lambda):
        assert str(lambda_node.inf_type.getValue()) == "typing.Callable[[~_T0, ~_T0], ~_T0]"
    for var_node in ast_mod.nodes_of_class(nodes.AssignName):
        if var_node.name == "y":
            y = ti.lookup_typevar(var_node, "y")
            assert ti.type_constraints.resolve(y).getValue() == str


def test_lambda_constant():
    src = """
    x = lambda: 1
    y = x()
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for var_node in ast_mod.nodes_of_class(nodes.AssignName):
        if var_node.name == "x":
            x = ti.lookup_typevar(var_node, "x")
            assert str(ti.type_constraints.resolve(x).getValue()) == "typing.Callable[[], int]"
        if var_node.name == "y":
            x = ti.lookup_typevar(var_node, "y")
            assert ti.type_constraints.resolve(x).getValue() == int


def test_nested_lambda():
    src = """
    f = lambda x: (lambda y: x + y)
    g = f(0)
    y = g(0)
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for var_node in ast_mod.nodes_of_class(nodes.AssignName):
        if var_node.name == "f":
            f = ti.lookup_typevar(var_node, "f")
            assert (
                str(ti.type_constraints.resolve(f).getValue())
                == "typing.Callable[[int], typing.Callable[[int], int]]"
            )
        if var_node.name == "g":
            g = ti.lookup_typevar(var_node, "g")
            assert str(ti.type_constraints.resolve(g).getValue()) == "typing.Callable[[int], int]"
        if var_node.name == "y":
            y = ti.lookup_typevar(var_node, "y")
            assert ti.type_constraints.resolve(y).getValue() == int



from typing import Callable, Type

from astroid import nodes
from hypothesis import settings

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


def test_builtin_function_name():
    """Test looking up the builtin function `bin`."""
    module, _ = cs._parse_text("bin")
    for node in module.nodes_of_class(nodes.Name):
        assert node.inf_type.getValue() == Callable[[int], str]


def test_builtin_class_name():
    """Test looking up the builtin class `int`."""
    module, _ = cs._parse_text("int")
    for node in module.nodes_of_class(nodes.Name):
        assert node.inf_type.getValue() == Type[int]



import typing

from astroid import nodes
from pytest import skip

from python_ta.typecheck.base import (
    TypeFail,
    TypeFailAnnotationUnify,
    TypeFailFunction,
    TypeFailUnify,
)

from .. import custom_hypothesis_support as cs


def find_type_fail(ast_node):
    if isinstance(ast_node.inf_type, TypeFail):
        return ast_node
    else:
        for child in ast_node.get_children():
            child_res = find_type_fail(child)
            if child_res is not None:
                return child_res
    return None


def verify_typefail_unify(tf: TypeFailUnify, *exp_tnodes, exp_src_type, num_reasons):
    assert isinstance(tf, TypeFailUnify)
    for tn, exp_r in zip(tf.tnodes, exp_tnodes):
        if tn.ast_node:
            assert tn.ast_node.name == exp_r
        else:
            assert tn.type == exp_r
    assert isinstance(tf.src_node, exp_src_type)

    reasons = []
    for tn in tf.tnodes:
        reasons += tn.find_path_to_parent()
    assert len(reasons) == num_reasons


def verify_typefail_function(tf: TypeFailFunction, act_func_call: str):
    assert isinstance(tf, TypeFailFunction)
    assert isinstance(tf.src_node, nodes.Call)
    assert tf.src_node.as_string() == act_func_call


def test_var_assign():
    src = """
    A = 1
    A = 'One'
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tf = find_type_fail(ast_mod).inf_type
    verify_typefail_unify(tf, "A", str, exp_src_type=nodes.Assign, num_reasons=1)


def test_two_var_assign():
    src = """
    A = 1
    B = 'Two'
    A = B
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tf = find_type_fail(ast_mod).inf_type
    verify_typefail_unify(tf, "A", "B", exp_src_type=nodes.Assign, num_reasons=2)


def test_var_chain():
    src = """
    A = 1
    Z = 'Zed'

    B = A
    C = B

    C = Z
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tf = find_type_fail(ast_mod).inf_type
    verify_typefail_unify(tf, "C", "Z", exp_src_type=nodes.Assign, num_reasons=4)


def test_one_list():
    src = """
    L1 = [1, 2, 3]
    L1 = "Hello"
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tf = find_type_fail(ast_mod).inf_type
    verify_typefail_unify(tf, "L1", str, exp_src_type=nodes.Assign, num_reasons=1)


def test_two_lists():
    src = """
    L1 = [1, 2, 3]
    L2 = ['a', 'b', 'c']
    L1 = L2
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tf = find_type_fail(ast_mod).inf_type
    verify_typefail_unify(tf, "L1", "L2", exp_src_type=nodes.Assign, num_reasons=2)


def test_tuple():
    src = """
    T1 = (1, 2)
    T2 = ('a', 'b')
    T1 = T2
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tf = find_type_fail(ast_mod).inf_type
    verify_typefail_unify(tf, "T1", "T2", exp_src_type=nodes.Assign, num_reasons=2)


def test_annotation():
    src = """
    x: int
    x = 'Hello'
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tf = find_type_fail(ast_mod).inf_type
    assert isinstance(tf, TypeFailAnnotationUnify)


def test_func_annotation():
    src = """
    def f(x: int):
        return x

    f('Hello')
    """
    skip("Requires modifications to unify_call")
    ast_mod, ti = cs._parse_text(src, reset=True)
    tf = find_type_fail(ast_mod).inf_type
    assert isinstance(tf, TypeFailAnnotationUnify)


def test_function():
    src = """
    def f(x):
        return x + 1

    f('Hello')
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tf = find_type_fail(ast_mod).inf_type
    verify_typefail_function(tf, "f('Hello')")


def test_function_index():
    src = """
    def f(x, y):
        return x + y + 1

    f(1, 'two')
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tf = find_type_fail(ast_mod).inf_type
    verify_typefail_function(tf, "f(1, 'two')")
    assert tf.arg_indices == [1]


def test_function_multi_index():
    src = """
    def f(x, y):
        return x + y + 1

    f('one', 'two')
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tf = find_type_fail(ast_mod).inf_type
    verify_typefail_function(tf, "f('one', 'two')")
    assert tf.arg_indices == [0, 1]


def test_function_numargs():
    src = """
    def f(x, y):
        return x + y

    f(5)
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tf = find_type_fail(ast_mod).inf_type
    verify_typefail_function(tf, "f(5)")


def test_function_overload():
    src = """
    def f(x, y=0):
        return x + y + 1

    f(1, 2, 3)"""
    ast_mod, ti = cs._parse_text(src, reset=True)
    tf = find_type_fail(ast_mod).inf_type
    verify_typefail_function(tf, "f(1, 2, 3)")



"""
Sanity checker to pretty print the abstract syntax tree (AST) structure.
"""
import sys

import astroid
from colorama import Back, Fore

USAGE = "Usage: python -m sample_usage.print_ast <your-file.py>"
FILL = "  "
# Unicode Box Drawing Characters.
CHAR_PIPE = "┃"
CHAR_TUBE = "━"


def walker(node, source_lines, indent=""):
    """Recursively visit the ast in a preorder traversal."""

    node_name = str(node)[0 : str(node).index("(")]
    value = None
    if hasattr(node, "value"):
        if "(" in str(node.value):
            value = str(node.value)[0 : str(node.value).index("(")]
        else:
            value = node.value

    name = node.name if hasattr(node, "name") else None
    print("{}{} {} (name: {}, value: {})".format(indent, CHAR_TUBE, node_name, name, value))

    lines = [line for line in node.as_string().split("\n")]
    for line in lines:
        print(indent + FILL + ">>>" + Fore.BLUE + line + Fore.RESET)

    for child in node.get_children():
        walker(child, source_lines, indent + FILL + CHAR_PIPE)


if __name__ == "__main__":
    if len(sys.argv) != 2:
        print(USAGE)
        exit()

    with open(sys.argv[1]) as file:
        file_contents = file.read()
        astroid_module = astroid.parse(file_contents)
        source_lines = file_contents.split("\n")
        print('Note: code in the ">>>" section is an approximation.\n')
        for node in astroid_module.body:
            walker(node, source_lines, CHAR_PIPE)
            print("┣" + CHAR_TUBE * 60 + "┫")
        print("")



import os
from collections import OrderedDict

import python_ta
from python_ta.reporters.stat_reporter import StatReporter
from sample_usage.stats_analysis import summary


def pyta_statistics(directory, config=""):
    """
    Recursively run python_ta.check_all() on the files in the directory and its
    subdirectories to collect the error and style messages.
    Assume directory is a folder that contains subdirectories named by
    student's UTORid/studentID/group name, or directory can also be a student
    folder itself. Also assume that student's directory doesn't contain any
    subdirectories, only files.

    @param str directory: The string of the path way of the directory.
    @param str config: A path to the configuration file to use.
    @rtype: dict[str, str]
    """
    all_errors = OrderedDict()

    for root_str, dir_list, file_list in os.walk(directory):
        # check if directory is student directory
        if len(dir_list) == 0:
            student_id = os.path.basename(root_str)

            for file in file_list:
                if file.endswith(".py"):
                    python_ta.check_all(
                        os.path.join(root_str, file),
                        reporter=StatReporter,
                        config=config,
                        number_of_messages=0,
                    )
                    # store all the msg objects of this student's files
            all_errors[student_id] = (StatReporter.error_messages, StatReporter.style_messages)

    if len(all_errors) == 0:
        raise Exception("No student files found in given directory!")
    else:
        _print_stats(*summary(all_errors))
        return all_errors


def _print_stats(individual_stats, summary_stats):
    """Pretty prints the statistics aggregated by summary.

    @param OrderedDict[str, List] individual_stats:
        stats computed per student/group
    @param OrderedDict[str, List | OrderedDict] summary_stats:
        stats computed over all the students
    @rtype: dict

    Format:
    ===================== Statistics per Student/Group =====================
    Student/group: <student/group id>
        Totals:
            Error and Style Messages: <number>
            Error Messages:           <number>
            Style Messages:           <number>

        Most Frequent Messages:
            1. <Error Type 1>: <number> (<percentage>%)
            2. ...
        Most Frequent Errors:
            1. ...
        Most Frequent Styles:
            1. ...

    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Student/group: ...
        ...

    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


    ================== Aggregate Statistics for Directory ==================
    Most Frequent Code Errors:
        1. <Error Type 1>: <number>
        2. ...
    Average Code Errors Per Student: <number>

    Most ...
        ...
    ...

    ...

    ======= Five Number Summary of Errors Per Student =======
        Most Errors:         <number>
        Upper Quartile (Q3): <number>
        Median:              <number>
        Lower Quartile (Q1): <number>
        Least Errors:        <number>

    ~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~=~
    """
    print("===================== Statistics per Submission", "=====================")

    for name, stats in individual_stats.items():
        print("Submission by:", name)

        # Totals are printed differently
        header, totals = stats.pop(0)
        print("\t{}:".format(header))
        for total, count in totals:
            print("\t\t{:26}{}".format(total + ":", count))

        # Print "most frequent" stats
        for stat_type, results in stats:
            print("\t{}:".format(stat_type))
            _print_top_errors(results, 2, False)
        print("\n~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~\n")

    if len(summary_stats) == 0:
        return

    print("\n================== Aggregate Statistics for Directory", "==================")

    # Most Frequent Code Errors
    stat_type, stats = summary_stats[0]
    print(stat_type + ":")
    _print_top_errors(stats)

    # Average Code Errors Per Student
    print("{}: {}\n".format(summary_stats[1][0], summary_stats[1][1]))

    # Most Frequent Style Errors
    stat_type, stats = summary_stats[2]
    print(stat_type + ":")
    _print_top_errors(stats)

    # Average Style Errors Per Student
    print("{}: {}\n".format(summary_stats[3][0], summary_stats[3][1]))

    # Most Frequent Errors of Either Type
    stat_type, stats = summary_stats[4]
    print(stat_type + ":")
    _print_top_errors(stats)

    # Average Errors of Either Type Per Student
    print("{}: {}\n".format(summary_stats[5][0], summary_stats[5][1]))

    # Five Number Summary
    print("======= {} =======".format(summary_stats[6][0]))
    five_numbers = summary_stats[6][1]
    for stat_type, value in five_numbers:
        print("\t{:21}{}".format(stat_type + ":", value))

    # Standard Deviation
    print("\n{}: {}".format(summary_stats[7][0], summary_stats[7][1]))


def _print_top_errors(stats, tabs=1, aggregate=True):
    """Pretty prints a list of most frequent errors and their counts.

    @param List[Tuple[str, int | Tuple]] stats: the errors & counts to print
    @param int tabs: the number of tabs to print before every line
    @param bool aggregate: whether stats are aggregate (i.e.: no percentages)
    @rtype: None
    """
    for i, stat in enumerate(stats):
        print("{}{}.".format("\t" * tabs, i + 1), end=" ")
        # TODO: figure out a non-hardcoded alternative for 38 in prints:
        if aggregate:
            print("{:38} {:3}".format(stat[0] + ":", stat[1]))
        else:
            print("{:38} {:3} ({}%)".format(stat[0][0] + ":", stat[0][1], stat[1][1]))



from keyword import iskeyword
from typing import Any, List, TypeVar

import hypothesis.strategies as hs
import pytest
from astroid import nodes
from hypothesis import HealthCheck, assume, given, settings
from pytest import skip

from python_ta.typecheck.base import TypeFail

from .. import custom_hypothesis_support as cs
from ..custom_hypothesis_support import lookup_type

settings.load_profile("pyta")


@given(cs.random_dict_variable_homogeneous_value(min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_set_env(variables_dict):
    """Test environment setting visitors"""
    program = cs._parse_dictionary_to_program(variables_dict)
    module, _ = cs._parse_text(program)
    # get list of variable names in locals
    local_values = [module.type_environment.locals[name] for name in module.type_environment.locals]
    global_values = [
        module.type_environment.globals[name] for name in module.type_environment.globals
    ]
    # verify the type of the value of each variable in the environment
    for value in local_values:
        assert isinstance(value, TypeVar)
    for value in global_values:
        assert isinstance(value, TypeVar)


@given(cs.valid_identifier())
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_set_name_unassigned(identifier):
    """Test visitor for name nodes representing a single unassigned variable in module."""
    program = identifier
    module, _ = cs._parse_text(program)
    for name_node in module.nodes_of_class(nodes.Name):
        assert isinstance(name_node.inf_type, TypeFail)


@given(cs.random_dict_variable_homogeneous_value(min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_set_name_assigned(variables_dict):
    """Test visitor for name nodes representing a variables with assigned values in module."""
    program = cs._parse_dictionary_to_program(variables_dict)
    for variable_name in variables_dict:
        program += variable_name + "\n"
    module, inferer = cs._parse_text(program)
    for name_node in module.nodes_of_class(nodes.Name):
        name_type = lookup_type(inferer, name_node, name_node.name)
        assert name_node.inf_type.getValue() == name_type


@given(cs.random_dict_variable_homogeneous_value(min_size=1))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_set_single_assign(variables_dict):
    """Test single-target assignment statements; verify unification of type variables."""
    program = cs._parse_dictionary_to_program(variables_dict)
    module, inferer = cs._parse_text(program)
    for node in module.nodes_of_class(nodes.AssignName):
        target_value = node.parent.value
        target_type = lookup_type(inferer, node, node.name)
        # compare it to the type of the assigned value
        assert target_value.inf_type.getValue() == target_type


@given(cs.random_dict_variable_homogeneous_value(min_size=2))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_multi_target_assign(variables_dict):
    """Test multi-target assignment statements; verify unification of type variables."""
    for variable_name in variables_dict:
        assume(not iskeyword(variable_name))
    program = (
        ", ".join(variables_dict.keys())
        + " = "
        + ", ".join([repr(value) for value in variables_dict.values()])
    )
    module, inferer = cs._parse_text(program)
    # for each Assign node in program, verify unification of the type variables.
    for node in module.nodes_of_class(nodes.Assign):
        target_type_tuple = zip(node.targets[0].elts, node.value.elts)
        for target, value in target_type_tuple:
            target_type_var = target.frame().type_environment.lookup_in_env(target.name)
            assert (
                inferer.type_constraints.resolve(target_type_var).getValue()
                == value.inf_type.getValue()
            )


@given(
    hs.lists(hs.text(alphabet="abcdefghijklmnopqrstuvwxyz", min_size=1), min_size=1),
    cs.primitive_values,
)
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_set_multi_assign(variables_list, value):
    """Test environment setting visitors"""
    for variable_name in variables_list:
        assume(not iskeyword(variable_name))
    variables_list.append(repr(value))
    program = (" = ").join(variables_list)
    module, inferer = cs._parse_text(program)
    for target_node in module.nodes_of_class(nodes.AssignName):
        value_type = target_node.parent.value.inf_type.getValue()
        target_type_var = target_node.frame().type_environment.lookup_in_env(target_node.name)
        assert inferer.type_constraints.resolve(target_type_var).getValue() == value_type


@given(cs.random_dict_variable_homogeneous_value(min_size=2))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_assign_complex_homogeneous(variables_dict):
    """test whether visitors properly set the type constraint of the a assign node representing a multi-target-assign
    with a homogeneous list as the value.
    """
    for variable_name in variables_dict:
        assume(not iskeyword(variable_name))
    program = (
        "x = ["
        + ", ".join([repr(value) for value in variables_dict.values()])
        + "]\n"
        + ", ".join(variables_dict.keys())
        + " = x"
    )
    module, typeinferrer = cs._parse_text(program)
    ass_node = list(module.nodes_of_class(nodes.Assign))[0]
    for variable_name in variables_dict:
        var_tvar = module.type_environment.lookup_in_env(variable_name)
        assert (
            typeinferrer.type_constraints.resolve(var_tvar).getValue()
            == ass_node.value.elts[0].inf_type.getValue()
        )


@pytest.mark.skip(reason="Not a test")
@given(hs.lists(cs.valid_identifier(), min_size=2), cs.random_list(min_size=2))
def test_assign_complex(variables, values):
    """Test whether visitors properly set the type constraint of the a Assign node representing a multi-target-assign
    with a homogeneous list as the value.
    """
    assume(type(values[0]) != type(values[1]) and len(variables) == len(values))
    val_types = [type(val) for val in values]
    if int in val_types:
        assume(bool not in val_types)
    if bool in val_types:
        assume(int not in val_types)
    program = (
        "x = [" + ", ".join([repr(val) for val in values]) + "]\n" + ", ".join(variables) + " = x"
    )
    module, TypeInferrer = cs._parse_text(program)
    for variable_name in variables:
        variable_type_var = module.type_environment.lookup_in_env(variable_name)
        assert TypeInferrer.type_constraints.resolve(variable_type_var).getValue() == Any


def test_attribute_reassign():
    """Test for correct type setting after a redundant assignment of an instance attribute."""
    program = (
        f"class Student:\n"
        f"    def __init__(self, name1):\n"
        f"        self.name = name1\n"
        f"        self.name = name1\n"
        f"\n"
    )
    module, inferer = cs._parse_text(program)
    functiondef_node = next(module.nodes_of_class(nodes.FunctionDef))
    actual_type = lookup_type(inferer, functiondef_node, "name")
    expected_type = lookup_type(inferer, functiondef_node, "name1")
    assert inferer.type_constraints.can_unify(actual_type, expected_type)


def test_assign_autoconvert():
    skip("TODO: make this test pass (currently a unification fail)")
    program = """
    x = 1
    x = x + 1.0
    """
    module, inferer = cs._parse_text(program, reset=True)
    x = [
        inferer.lookup_typevar(node, node.name) for node in module.nodes_of_class(nodes.AssignName)
    ][0]
    assert inferer.type_constraints.resolve(x).getValue() == float


def test_augassign_simple_fallback():
    program = """
    x += 1
    """
    module, inferer = cs._parse_text(program, reset=True)
    x = [
        inferer.lookup_typevar(node, node.name) for node in module.nodes_of_class(nodes.AssignName)
    ][0]
    assert inferer.type_constraints.resolve(x).getValue() == int


def test_augassign_subscript_target():
    program = """
    x = ['a', 'b', 'c']
    x[0] += 'd'
    y = x[0]
    """
    module, inferer = cs._parse_text(program, reset=True)
    x, y = [
        inferer.lookup_typevar(node, node.name) for node in module.nodes_of_class(nodes.AssignName)
    ]
    assert inferer.type_constraints.resolve(y).getValue() == str


def test_augassign_builtin():
    program = """
    x = [1,2,3]
    x += set([4,5,6])
    """
    module, inferer = cs._parse_text(program, reset=True)
    x = [
        inferer.lookup_typevar(node, node.name) for node in module.nodes_of_class(nodes.AssignName)
    ][0]
    assert inferer.type_constraints.resolve(x).getValue() == List[int]


def test_augassign_builtin_autoconvert():
    skip("TODO: make this test pass (currently a unification fail)")
    program = """
    x = 1
    x += 1.0
    """
    module, inferer = cs._parse_text(program, reset=True)
    x = [
        inferer.lookup_typevar(node, node.name) for node in module.nodes_of_class(nodes.AssignName)
    ][0]
    assert inferer.type_constraints.resolve(x).getValue() == float


def test_augassign_builtin_attribute_autoconvert():
    skip("TODO: make this test pass (currently a unification fail)")
    program = """
    class A:
        def __init__(self):
            self.x = 1

    self.x += 1.0
    """
    module, inferer = cs._parse_text(program, reset=True)
    x = [
        inferer.lookup_typevar(node, node.name) for node in module.nodes_of_class(nodes.AssignAttr)
    ][0]
    assert inferer.type_constraints.resolve(x).getValue() == float


def test_augassign_userdefn():
    skip("Support for attribute-based inference required for this test to pass")
    program = """
    class A:
        def __init__(self, val):
            self.val = val

        def __iadd__(self, other):
            return A(int(self.val + other.val))

        def __add__(self, other):
            return A(str(self.val + other.val))

    a = A(0)
    b = A(1)
    a += b
    x = a.val
    """
    module, inferer = cs._parse_text(program, reset=True)
    a, b, _, x = [
        inferer.lookup_typevar(node, node.name) for node in module.nodes_of_class(nodes.AssignName)
    ][6:]
    assert inferer.type_constraints.resolve(x).getValue() == int


def test_augassign_userdefn_fallback():
    skip("Support for attribute-based inference required for this test to pass")
    program = """
    class A:
        def __init__(self, val):
            self.val = val

        def __add__(self, other):
            return A(int(self.val + other.val))

    a = A(0)
    b = A(1)
    a += b
    x = a.val
    """
    module, inferer = cs._parse_text(program, reset=True)
    a, b, _, x = [
        inferer.lookup_typevar(node, node.name) for node in module.nodes_of_class(nodes.AssignName)
    ][4:]
    assert inferer.type_constraints.resolve(x).getValue() == int


def test_augassign_userdefn_fail():
    program = """
    class A:
        def __init__(self, val):
            self.val = val

    a = A(0)
    b = A(1)
    a += b
    x = a.val
    """
    module, inferer = cs._parse_text(program, reset=True)
    for aug_node in module.nodes_of_class(nodes.AugAssign):
        assert isinstance(aug_node.inf_type, TypeFail)



from ast import literal_eval

from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker


class InvalidRangeIndexChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = "invalid_range_index"
    msgs = {
        "E9993": (
            "You should not use invalid range index on line %s",
            "invalid-range-index",
            "Used when you use invalid index range",
        )
    }
    # this is important so that your checker is executed before others
    priority = -1

    @check_messages("invalid-range-index")
    def visit_call(self, node):
        if isinstance(node.func, nodes.Name):
            name = node.func.name
            # ignore the name if it's not a builtin (i.e. not defined in the
            # locals nor globals scope)
            if not (name in node.frame() or name in node.root()) and name == "range":
                arg = node.args  # the arguments of 'range' call
                # guard nodes (e.g. Name) not properly handled by literal_eval.
                if any([not isinstance(item, nodes.Const) for item in arg]):
                    return
                eval_parm = list(map(lambda z: literal_eval(z.as_string()), arg))

                # check if there is no args in 'range' call
                if (
                    len(arg) == 0
                    or not all([isinstance(c, int) for c in eval_parm])
                    or (len(arg) == 1 and eval_parm[0] < 2)
                    or (len(arg) == 2 and eval_parm[1] - eval_parm[0] < 2)
                ):

                    args = "{}".format(node.lineno)
                    self.add_message("invalid-range-index", node=node, args=args)

                if len(arg) == 3:
                    if (
                        abs(eval_parm[2]) >= abs(eval_parm[0] - eval_parm[1])
                        or eval_parm[2] == 0
                        or (eval_parm[0] > eval_parm[1] and eval_parm[2] < 0)
                        or (eval_parm[0] < eval_parm[1] and eval_parm[2] > 0)
                    ):

                        args = "{}".format(node.lineno)
                        self.add_message("invalid-range-index", node=node, args=args)


def register(linter):
    """required method to auto register this checker"""
    linter.register_checker(InvalidRangeIndexChecker(linter))



import typing
from itertools import chain
from typing import *
from typing import Callable, ForwardRef, Union, _GenericAlias

import astroid
import astroid.inference
from astroid import nodes
from astroid.transforms import TransformVisitor

from ..typecheck.base import (
    Environment,
    NoType,
    TypeConstraints,
    TypeFail,
    TypeFailAnnotationInvalid,
    TypeFailFunction,
    TypeFailLookup,
    TypeFailReturn,
    TypeFailStarred,
    TypeInfo,
    TypeResult,
    _ann_node_to_type,
    _gorg,
    _node_to_type,
    accept_failable,
    create_Callable_TypeResult,
    failable_collect,
    is_callable,
    wrap_container,
)
from ..typecheck.errors import (
    BINOP_TO_METHOD,
    BINOP_TO_REV_METHOD,
    INPLACE_TO_BINOP,
    UNARY_TO_METHOD,
    binop_error_message,
    subscript_error_message,
    unaryop_error_message,
)
from ..typecheck.type_store import TypeStore


class TypeInferer:
    """The class responsible for inferring types given an astroid AST."""

    type_constraints = TypeConstraints()
    type_store = TypeStore(type_constraints)
    type_constraints.type_store = type_store

    def __init__(self) -> None:
        self.type_constraints.reset()

    def reset(self) -> None:
        self.type_constraints.reset()
        self.type_store = TypeStore(self.type_constraints)
        self.type_constraints.type_store = self.type_store

    ###########################################################################
    # Setting up the environment
    ###########################################################################
    def environment_transformer(self) -> TransformVisitor:
        """Return a TransformVisitor that sets an environment for every node."""
        visitor = TransformVisitor()
        visitor.register_transform(nodes.FunctionDef, self._set_function_def_environment)
        visitor.register_transform(nodes.AsyncFunctionDef, self._set_function_def_environment)
        visitor.register_transform(nodes.ClassDef, self._set_classdef_environment)
        visitor.register_transform(nodes.Module, self._set_module_environment)
        visitor.register_transform(nodes.ListComp, self._set_comprehension_environment)
        visitor.register_transform(nodes.DictComp, self._set_comprehension_environment)
        visitor.register_transform(nodes.SetComp, self._set_comprehension_environment)
        visitor.register_transform(nodes.GeneratorExp, self._set_comprehension_environment)
        visitor.register_transform(nodes.Lambda, self._set_comprehension_environment)
        return visitor

    def _set_module_environment(self, node: nodes.Module) -> None:
        """Method to set environment of a Module node."""
        node.type_environment = Environment()
        for name in node.globals:
            if not any(
                isinstance(elt, (nodes.ImportFrom, nodes.Import)) for elt in node.globals[name]
            ):
                new_tvar = self.type_constraints.fresh_tvar(node.globals[name][0])
                if any(isinstance(elt, nodes.ClassDef) for elt in node.globals[name]):
                    self.type_constraints.unify(new_tvar, Type[ForwardRef(name)], node)
                node.type_environment.globals[name] = new_tvar
        self._populate_local_env(node)

    def _set_classdef_environment(self, node: nodes.ClassDef) -> None:
        """Method to set environment of a ClassDef node."""
        node.type_environment = Environment()
        for name in node.instance_attrs:
            node.type_environment.locals[name] = self.type_constraints.fresh_tvar(
                node.instance_attrs[name][0]
            )
            self.type_store.classes[node.name][name] = [
                (node.type_environment.locals[name], "attribute")
            ]
        for name in node.locals:
            if name in ["__module__", "__qualname__"]:
                node.type_environment.locals[name] = str
            else:
                node.type_environment.locals[name] = self.type_constraints.fresh_tvar(
                    node.locals[name][0]
                )

        self.type_store.classes[node.name]["__bases"] = [_node_to_type(base) for base in node.bases]
        try:
            self.type_store.classes[node.name]["__mro"] = [cls.name for cls in node.mro()]
        except astroid.exceptions.DuplicateBasesError:
            self.type_store.classes[node.name]["__mro"] = [node.name]

    def _set_function_def_environment(self, node: nodes.FunctionDef) -> None:
        """Method to set environment of a FunctionDef node."""
        node.type_environment = Environment()
        # self is a special case
        if (
            node.args.args
            and node.args.args[0].name == "self"
            and isinstance(node.parent, nodes.ClassDef)
        ):
            node.type_environment.locals["self"] = ForwardRef(node.parent.name)
        self._populate_local_env(node)
        self._populate_local_env_attrs(node)
        node.type_environment.locals["return"] = self.type_constraints.fresh_tvar(node)

    def _set_comprehension_environment(self, node: nodes.Comprehension) -> None:
        """Set the environment of a comprehension expression.

        Covers ListComp, SetComp, DictComp, and GeneratorExp."""
        node.type_environment = Environment()
        for name in node.locals:
            node.type_environment.locals[name] = self.type_constraints.fresh_tvar(node)

    def _populate_local_env(self, node: nodes.NodeNG) -> None:
        """Helper to populate locals attributes in type environment of given node."""
        for var_name in node.locals:
            try:
                var_value = node.type_environment.lookup_in_env(var_name)
            except KeyError:
                if any(
                    isinstance(elt, (nodes.ImportFrom, nodes.Import))
                    for elt in node.locals[var_name]
                ):
                    var_value = Any
                else:
                    var_value = self.type_constraints.fresh_tvar(node.locals[var_name][0])
            node.type_environment.locals[var_name] = var_value

    def _populate_local_env_attrs(self, node: nodes.NodeNG) -> None:
        """Store in TypeStore the attributes of any unresolved class names"""
        for attr_node in chain(
            node.nodes_of_class(nodes.Attribute), node.nodes_of_class(nodes.AssignAttr)
        ):
            if (
                isinstance(attr_node.expr, nodes.Name)
                and attr_node.expr.name in node.type_environment.locals
            ):
                class_type = node.type_environment.lookup_in_env(attr_node.expr.name)
                if isinstance(class_type, TypeVar):
                    self.type_store.classes[class_type.__name__]["__mro"] = [class_type.__name__]
                    if not attr_node.attrname in self.type_store.classes[class_type.__name__]:
                        self.type_store.classes[class_type.__name__][attr_node.attrname] = [
                            (self.type_constraints.fresh_tvar(attr_node), "attribute")
                        ]

    ###########################################################################
    # Type inference methods
    ###########################################################################
    def type_inference_transformer(self) -> TransformVisitor:
        """Instantiate a visitor to perform type inference on an AST."""
        type_visitor = TransformVisitor()
        for klass in nodes.ALL_NODE_CLASSES:
            if hasattr(self, f"visit_{klass.__name__.lower()}"):
                type_visitor.register_transform(
                    klass, getattr(self, f"visit_{klass.__name__.lower()}")
                )
            else:
                type_visitor.register_transform(klass, self.visit_default)
        return type_visitor

    def visit_default(self, node: nodes.NodeNG) -> None:
        node.inf_type = NoType()

    ##############################################################################
    # Literals
    ##############################################################################
    def visit_const(self, node: nodes.Const) -> None:
        node.inf_type = TypeInfo(type(node.value))

    def visit_list(self, node: nodes.List) -> None:
        if node.ctx == nodes.Store:
            # List is the target of an assignment; do not give it a type.
            node.inf_type = NoType()
        elif not node.elts:
            node.inf_type = TypeInfo(List[self.type_constraints.fresh_tvar(node)])
        else:
            elt_inf_type = self._unify_elements(node.elts, node)
            node.inf_type = wrap_container(List, elt_inf_type)

    def visit_set(self, node: nodes.Set) -> None:
        if not node.elts:
            node.inf_type = TypeInfo(Set[self.type_constraints.fresh_tvar(node)])
        else:
            elt_inf_type = self._unify_elements(node.elts, node)
            node.inf_type = wrap_container(Set, elt_inf_type)

    def visit_dict(self, node: nodes.Dict) -> None:
        if not node.items:
            node.inf_type = TypeInfo(
                Dict[self.type_constraints.fresh_tvar(node), self.type_constraints.fresh_tvar(node)]
            )
        else:
            key_list, val_list = zip(*node.items)
            key_inf_type = self._unify_elements(key_list, node)
            val_inf_type = self._unify_elements(val_list, node)
            node.inf_type = wrap_container(Dict, key_inf_type, val_inf_type)

    def visit_tuple(self, node: nodes.Tuple) -> None:
        if node.ctx == nodes.Store:
            # Tuple is the target of an assignment; do not give it a type.
            node.inf_type = NoType()
        else:
            node.inf_type = wrap_container(Tuple, *(e.inf_type for e in node.elts))

    def _unify_elements(self, lst: List[nodes.NodeNG], node: nodes.NodeNG) -> TypeResult:
        lst = list(lst)
        elt_inf_type = lst[0].inf_type
        for cur_elt in lst[1:]:
            elt_inf_type = self.type_constraints.unify(elt_inf_type, cur_elt.inf_type, node)
            if isinstance(elt_inf_type, TypeFail):
                return TypeInfo(Any)

        return elt_inf_type

    ##############################################################################
    # Expression types
    ##############################################################################
    def visit_ifexp(self, node: nodes.IfExp) -> None:
        node.inf_type = self.type_constraints.unify(node.body.inf_type, node.orelse.inf_type, node)

    def visit_expr(self, node: nodes.Expr) -> None:
        """Expr nodes take the type of their child."""
        node.inf_type = node.value.inf_type

    ##############################################################################
    # Name lookup and assignment
    ##############################################################################
    def visit_name(self, node: nodes.Name) -> None:
        node.inf_type = self.lookup_inf_type(node, node.name)

    def visit_assign(self, node: nodes.Assign) -> None:
        """Update the enclosing scope's type environment for the assignment's binding(s)."""
        # the type of the expression being assigned
        if isinstance(node.value, nodes.Name):
            expr_inf_type = self.lookup_typevar(node, node.value.name)
        else:
            expr_inf_type = node.value.inf_type

        node.inf_type = NoType()
        for target in node.targets:
            type_result = self._assign_type(target, expr_inf_type, node)
            if isinstance(type_result, TypeFail):
                node.inf_type = type_result
                break

    def visit_annassign(self, node: nodes.AnnAssign) -> None:
        if isinstance(node.target, nodes.AssignAttr):
            var_inf_type = self.lookup_typevar(node.target, node.target.attrname)
        else:
            var_inf_type = self.lookup_typevar(node.target, node.target.name)
        ann_type = _ann_node_to_type(node.annotation)
        self.type_constraints.unify(var_inf_type, ann_type, node)
        if node.value:
            node.targets = [node.target]
            self.visit_assign(node)
        elif isinstance(ann_type, TypeFail):
            node.inf_type = ann_type
        else:
            node.inf_type = NoType()

    def visit_augassign(self, node: nodes.AugAssign) -> None:
        node.inf_type = NoType()

        # lookup method for augmented arithmetic assignment
        method_name = BINOP_TO_METHOD[node.op]
        if isinstance(node.target, nodes.Subscript):
            target_type = node.target.value.inf_type
            binop_result = self._handle_call(
                node.target,
                "__setitem__",
                target_type,
                node.target.slice.inf_type,
                node.value.inf_type,
            )
        else:
            if isinstance(node.target, nodes.AssignName):
                target_type = self.lookup_typevar(node.target, node.target.name)
            elif isinstance(node.target, nodes.AssignAttr):
                target_type = self._lookup_attribute_type(
                    node.target, node.target.expr.inf_type, node.target.attrname
                )
            binop_result = self._handle_call(node, method_name, target_type, node.value.inf_type)
        if isinstance(binop_result, TypeFail):
            # on failure, fallback to method corresponding to standard operator
            boolop = INPLACE_TO_BINOP[node.op]
            method_name = BINOP_TO_METHOD[boolop]
            arithm_type = self._arithm_convert(node, method_name, target_type, node.value.inf_type)
            if arithm_type:
                binop_result = arithm_type
            else:
                binop_result = self._handle_call(
                    node, method_name, target_type, node.value.inf_type
                )

        type_result = self._assign_type(node.target, binop_result, node)
        if isinstance(type_result, TypeFail):
            node.inf_type = type_result

    @accept_failable
    def _assign_type(self, target: nodes.NodeNG, expr_type: type, node: nodes.Assign) -> TypeResult:
        """Update the type environment so that the target is bound to the given type."""
        if isinstance(target, nodes.AssignName):
            # A single identifier, e.g. x = ...
            target_type_var = self.lookup_typevar(target, target.name)
            return self.type_constraints.unify(target_type_var, expr_type, node)
        elif isinstance(target, nodes.AssignAttr):
            # Attribute mutation, e.g. x.y = ...
            attr_type = self._lookup_attribute_type(target, target.expr.inf_type, target.attrname)
            return self.type_constraints.unify(attr_type, expr_type, node)
        elif isinstance(target, nodes.Tuple):
            # Unpacking assignment, e.g. x, y = ...
            if getattr(expr_type, "__origin__", None) is tuple:
                assign_result = self._assign_tuple(target, expr_type, node)
            else:
                assign_result = self._handle_call(target, "__iter__", expr_type)

                target_tvars = self._get_tuple_targets(target)
                starred_target_found = False
                for tvar, elt in zip(target_tvars, target.elts):
                    if isinstance(elt, nodes.Starred) and not starred_target_found:
                        starred_target_found = True
                        unif_result = assign_result >> (
                            lambda t: self.type_constraints.unify(tvar, List[t.__args__[0]], node)
                        )
                    elif isinstance(elt, nodes.Starred) and starred_target_found:
                        unif_result = TypeFailStarred(node)
                    else:
                        unif_result = assign_result >> (
                            lambda t: self.type_constraints.unify(tvar, t.__args__[0], node)
                        )

                    if isinstance(unif_result, TypeFail):
                        return unif_result
            return assign_result
        elif isinstance(target, nodes.Subscript):
            # TODO: previous case must recursively handle this one
            return self._handle_call(
                target, "__setitem__", target.value.inf_type, target.slice.inf_type, expr_type
            )

    def _assign_tuple(self, target: nodes.Tuple, value: Any, node: nodes.Assign) -> TypeResult:
        """Unify tuple of type variables and tuple of types, within context of Assign statement."""
        starred_index = None
        for i in range(len(target.elts)):
            if isinstance(target.elts[i], nodes.Starred):
                if starred_index is None:
                    starred_index = i
                else:
                    return TypeFailStarred(node)

        target_tvars = self._get_tuple_targets(target)

        if starred_index is not None:
            starred_length = len(value.__args__) - len(target.elts) + 1
            starred_subvalues = node.value.elts[starred_index : starred_index + starred_length]
            starred_value = wrap_container(List, self._unify_elements(starred_subvalues, node))

            starred_target_tvar = target_tvars[starred_index]

            unif_result = self.type_constraints.unify(starred_target_tvar, starred_value, node)
            if isinstance(unif_result, TypeFail):
                return unif_result

            nonstarred_values = Tuple[
                value.__args__[:starred_index] + value.__args__[starred_index + starred_length :]
            ]
            nonstarred_targets = target_tvars
            nonstarred_targets.remove(nonstarred_targets[starred_index])

        else:
            nonstarred_values = value
            nonstarred_targets = target_tvars

        nonstarred_target_tuple = wrap_container(Tuple, *nonstarred_targets)

        unif_result = self.type_constraints.unify(nonstarred_target_tuple, nonstarred_values, node)
        if isinstance(unif_result, TypeFail):
            return unif_result

        assign_result = TypeInfo(value)
        return assign_result

    def _get_tuple_targets(self, t: nodes.Tuple) -> List[type]:
        target_tvars = []
        for subtarget in t.elts:
            if isinstance(subtarget, nodes.AssignAttr):
                target_tvars.append(
                    self._lookup_attribute_type(
                        subtarget, subtarget.expr.inf_type, subtarget.attrname
                    )
                )
            elif isinstance(subtarget, nodes.Starred):
                if isinstance(subtarget.value, nodes.AssignAttr):
                    target_tvars.append(
                        self.lookup_typevar(subtarget.value, subtarget.value.attrname)
                    )
                else:
                    target_tvars.append(self.lookup_typevar(subtarget.value, subtarget.value.name))
            elif isinstance(subtarget, nodes.Subscript):
                target_tvars.append(
                    self._handle_call(
                        subtarget, "__getitem__", subtarget.value.inf_type, subtarget.slice.inf_type
                    )
                )
            else:
                target_tvars.append(self.lookup_typevar(subtarget, subtarget.name))
        return target_tvars

    @accept_failable
    def _lookup_attribute_type(
        self, node: nodes.NodeNG, class_type: type, attribute_name: str
    ) -> TypeResult:
        """Given the node, class and attribute name, return the type of the attribute."""
        class_type = self.type_constraints.resolve(class_type)
        class_name, _, _ = self.get_attribute_class(class_type)
        if (
            class_name in self.type_store.classes
            and attribute_name in self.type_store.classes[class_name]
        ):
            return self.type_constraints.resolve(
                self.type_store.classes[class_name][attribute_name][0][0]
            )
        closest_frame = node.scope().lookup(class_name)[0]
        try:
            class_env = closest_frame.locals[class_name][0].type_environment
            result = self.type_constraints.resolve(class_env.lookup_in_env(attribute_name))
        except (KeyError, AttributeError):
            result = TypeFailLookup(self.type_constraints.get_tnode(class_type), node, node.parent)
        return result

    def lookup_typevar(self, node: nodes.NodeNG, name: str) -> TypeResult:
        """Given a variable name, return the equivalent TypeVar in the closest scope relative to given node."""
        cur_node = node

        while cur_node is not None:
            # Get first parent node with scope
            cur_scope = cur_node.scope()
            try:
                # Attempt to look up variable in type environment
                return TypeInfo(cur_scope.type_environment.lookup_in_env(name))
            except KeyError:
                # Variable not found in scope of current node, search parent node
                cur_node = cur_scope.parent

        # If root of astroid tree is reached with no variable found,
        # search builtins and TypeStore for variable type
        if name in self.type_store.classes:
            result = TypeInfo(Type[__builtins__[name]])
        elif name.lower() in self.type_store.classes:
            result = TypeInfo(Type[__builtins__[name.lower()]])
        elif name in self.type_store.functions:
            result = TypeInfo(
                Union[tuple([func_type for func_type, _ in self.type_store.functions[name]])]
            )
        else:
            result = TypeFail("Unbound identifier")

        return result

    def lookup_inf_type(self, node: nodes.NodeNG, name: str) -> TypeResult:
        """Given a variable name, return a TypeResult object containing the type in the closest scope relative to given node."""
        tvar = self.lookup_typevar(node, name)
        return self.type_constraints.resolve(tvar)

    ##############################################################################
    # Operation nodes
    ##############################################################################
    @accept_failable
    def get_call_signature(self, c: type, node: nodes.NodeNG) -> TypeResult:
        """Check for and return initializer function signature when using class name as Callable.
        Return Callable unmodified otherwise.

        :param c: Class, ForwardRef to a class, or Callable
        :param node: nodes.Call node where function call is occurring
        """
        # Any is interpreted as a function that can take any arguments.
        if c is Any:
            return TypeInfo(Callable[..., Any])
        # Callable type; e.g., 'Callable[[int], int]'
        elif is_callable(c):
            return TypeInfo(c)
        # Union of Callables
        elif getattr(c, "__origin__", None) is Union and all(
            is_callable(elt) for elt in c.__args__
        ):
            return TypeInfo(c)
        # Class types; e.g., 'Type[ForwardRef('A')]'
        elif getattr(c, "__origin__", None) is type:
            class_type = c.__args__[0]
            if isinstance(class_type, ForwardRef):
                class_name = c.__args__[0].__forward_arg__
            else:
                class_name = class_type.__name__

            if "__init__" in self.type_store.classes[class_name]:
                matching_init_funcs = []
                for func_type, _ in self.type_store.classes[class_name]["__init__"]:
                    new_func_type = Callable[list(func_type.__args__[1:-1]), func_type.__args__[0]]
                    matching_init_funcs.append(new_func_type)
                init_func = Union[tuple(matching_init_funcs)]
            else:
                # Classes declared without initializer
                init_func = Callable[[], class_type]
            return TypeInfo(init_func)
        # Class instances; e.g., 'ForwardRef('A')'
        elif isinstance(c, ForwardRef):
            class_type = c
            class_name = c.__forward_arg__

            if "__call__" in self.type_store.classes[class_name]:
                call_args = list(self.type_store.classes[class_name]["__call__"][0][0].__args__)
                call_func = Callable[call_args[1:-1], call_args[-1]]
                return TypeInfo(call_func)
            else:
                class_tnode = self.type_constraints.get_tnode(class_type)
                return TypeFailLookup(class_tnode, node, node.parent)
        else:
            return TypeFailFunction((c,), None, node)

    def visit_call(self, node: nodes.Call) -> None:
        f = self.type_constraints.resolve(node.func.inf_type)
        func_inf_type = self.get_call_signature(f, node.func)
        arg_inf_types = [arg.inf_type for arg in node.args]

        node.inf_type = self.type_constraints.unify_call(func_inf_type, *arg_inf_types, node=node)

    def visit_binop(self, node: nodes.BinOp) -> None:
        left_inf, right_inf = node.left.inf_type, node.right.inf_type

        method_name = BINOP_TO_METHOD[node.op]
        # attempt to obtain a common arithmetic type
        arithm_type = self._arithm_convert(node, method_name, left_inf, right_inf)
        if arithm_type:
            node.inf_type = arithm_type
        else:
            rev_method_name = BINOP_TO_REV_METHOD[node.op]
            l_type = self._handle_call(node, method_name, left_inf, right_inf)
            r_type = self._handle_call(node, rev_method_name, right_inf, left_inf)

            if self.type_store.is_descendant(right_inf.getValue(), left_inf.getValue()):
                if isinstance(r_type, TypeFail) and isinstance(l_type, TypeInfo):
                    node.inf_type = l_type
                else:
                    node.inf_type = r_type
            else:
                if isinstance(l_type, TypeFail) and isinstance(r_type, TypeInfo):
                    node.inf_type = r_type
                else:
                    node.inf_type = l_type

    @accept_failable
    def _arithm_convert(
        self, node: nodes.NodeNG, method: str, t1: type, t2: type
    ) -> Optional[TypeInfo]:
        if t1 is complex and t2 is complex:
            common_type = complex
        elif (t1 is complex and issubclass(t2, typing.SupportsFloat)) or (
            t2 is complex and issubclass(t1, typing.SupportsFloat)
        ):
            # TODO: handle complex better. Looks like int, float don't
            # support typing.SupportsComplex.
            common_type = complex
        elif (t1 is float and issubclass(t2, typing.SupportsFloat)) or (
            t2 is float and issubclass(t1, typing.SupportsFloat)
        ):
            common_type = float
        else:
            common_type = None

        if common_type:
            return self._handle_call(node, method, common_type, common_type)
        else:
            return None

    def visit_unaryop(self, node: nodes.UnaryOp) -> None:
        # 'not' is not a function, so this handled as a separate case.
        if node.op == "not":
            node.inf_type = TypeInfo(bool)
        else:
            method_name = UNARY_TO_METHOD[node.op]
            node.inf_type = self._handle_call(node, method_name, node.operand.inf_type)

    def visit_boolop(self, node: nodes.BoolOp) -> None:
        node.inf_type = self._unify_elements(node.values, node)
        if isinstance(node.inf_type, TypeFail):
            node.inf_type = TypeInfo(Any)

    def _handle_compare(
        self, node: nodes.NodeNG, comparator: str, left: nodes.NodeNG, right: nodes.NodeNG
    ) -> TypeResult:
        """Helper function to lookup a comparator, find the equivalent function call,
        and unify call with given arguments.
        """
        if comparator == "is" or comparator == "is not":
            return TypeInfo(bool)
        elif comparator == "in" or comparator == "not in":
            return self._handle_call(
                node, BINOP_TO_METHOD[comparator], right.inf_type, left.inf_type
            )
        else:
            return self._handle_call(
                node, BINOP_TO_METHOD[comparator], left.inf_type, right.inf_type
            )

    def visit_compare(self, node: nodes.Compare) -> None:
        left = node.left
        compare_type = self._handle_compare(node, node.ops[0][0], left, node.ops[0][1])

        for comparator, right in node.ops[1:]:
            resolved_type = self._handle_compare(node, comparator, left, right)
            compare_type = self.type_constraints.unify(compare_type, resolved_type, node)

        node.inf_type = compare_type

    ##############################################################################
    # Subscripting
    ##############################################################################
    def visit_index(self, node: nodes.Index) -> None:
        node.inf_type = node.value.inf_type

    def visit_slice(self, node: nodes.Slice) -> None:
        lower_type = node.lower.inf_type if node.lower else type(None)
        upper_type = node.upper.inf_type if node.upper else type(None)
        step_type = node.step.inf_type if node.step else type(None)
        node.inf_type = self._handle_call(
            node, "__init__", slice, lower_type, upper_type, step_type
        )
        node.inf_type = node.inf_type >> (
            lambda t: TypeInfo(slice) if t == type(None) else TypeInfo(t)
        )

    def visit_extslice(self, node: nodes.ExtSlice):
        unif_res = failable_collect(dim.inf_type for dim in node.dims)
        node.inf_type = unif_res >> (lambda lst: wrap_container(Tuple, *lst))

    def visit_subscript(self, node: nodes.Subscript) -> None:
        if isinstance(node.slice.inf_type, TypeFail):
            node.inf_type = node.slice.inf_type
        elif node.ctx == nodes.Load:
            try:
                val_inf_type = self.type_constraints.resolve(node.value.inf_type)
                value_gorg = val_inf_type >> _gorg
            except AttributeError:
                value_gorg = None

            if value_gorg is type and isinstance(node.slice, nodes.Index):
                if isinstance(node.slice.value, nodes.Tuple):
                    node.inf_type = wrap_container(
                        _node_to_type(node.value), *_node_to_type(node.slice.value)
                    )
                else:
                    node.inf_type = wrap_container(
                        _node_to_type(node.value), _node_to_type(node.slice.value)
                    )
            else:
                node.inf_type = self._handle_call(
                    node, "__getitem__", node.value.inf_type, node.slice.inf_type
                )
        elif node.ctx == nodes.Store:
            node.inf_type = NoType()
        elif node.ctx == nodes.Del:
            node.inf_type = self._handle_call(
                node, "__delitem__", node.value.inf_type, node.slice.inf_type
            )

    ##############################################################################
    # Loops
    ##############################################################################
    def visit_for(self, node: Union[nodes.For, nodes.Comprehension]) -> None:
        iter_type_result = self._handle_call(node, "__iter__", node.iter.inf_type)
        if isinstance(node.target, nodes.AssignName):
            target_inf_type = self.lookup_inf_type(node.target, node.target.name)
        elif isinstance(node.target, nodes.AssignAttr):
            target_inf_type = self._lookup_attribute_type(
                node.target, node.target.expr.inf_type, node.target.attrname
            )
        elif isinstance(node.target, nodes.Subscript):
            target_inf_type = iter_type_result >> (
                lambda t: self._handle_call(
                    node.target,
                    "__setitem__",
                    node.target.value.inf_type,
                    node.target.slice.inf_type,
                    t.__args__[0],
                )
            )
        elif isinstance(node.target, nodes.Tuple):
            target_inf_type = wrap_container(
                Tuple,
                *[
                    self.lookup_inf_type(subtarget, subtarget.name)
                    for subtarget in node.target.elts
                ],
            )
        iter_type_result >> (
            lambda t: self.type_constraints.unify(t.__args__[0], target_inf_type, node)
        )
        node.inf_type = iter_type_result if isinstance(iter_type_result, TypeFail) else NoType()

    ##############################################################################
    # Comprehensions
    ##############################################################################
    def visit_comprehension(self, node: nodes.Comprehension) -> None:
        self.visit_for(node)

    def visit_dictcomp(self, node: nodes.DictComp) -> None:
        key_inf_type = self.type_constraints.resolve(node.key.inf_type)
        val_inf_type = self.type_constraints.resolve(node.value.inf_type)
        node.inf_type = wrap_container(Dict, key_inf_type, val_inf_type)

    def visit_generatorexp(self, node: nodes.GeneratorExp) -> None:
        elt_inf_type = self.type_constraints.resolve(node.elt.inf_type)
        node.inf_type = wrap_container(Generator, elt_inf_type, None, None)

    def visit_listcomp(self, node: nodes.ListComp) -> None:
        val_inf_type = self.type_constraints.resolve(node.elt.inf_type)
        node.inf_type = wrap_container(List, val_inf_type)

    def visit_setcomp(self, node: nodes.SetComp) -> None:
        elt_inf_type = self.type_constraints.resolve(node.elt.inf_type)
        node.inf_type = wrap_container(Set, elt_inf_type)

    @accept_failable
    def _handle_call(self, node: nodes.NodeNG, function_name: str, *arg_types: type) -> TypeResult:
        """Helper to lookup a function and unify it with given arguments.
        Return the return type of unified function call.
        """
        arg_inf_types = [self.type_constraints.resolve(arg) for arg in arg_types]
        func_type = self.type_store.lookup_method(function_name, *arg_inf_types, node=node)

        return self.type_constraints.unify_call(func_type, *arg_types, node=node)

    ##############################################################################
    # Definitions
    ##############################################################################
    def visit_functiondef(self, node: nodes.FunctionDef) -> None:
        node.inf_type = NoType()

        # Get the inferred type of the function arguments
        inferred_args = [self.lookup_inf_type(node, arg) for arg in node.argnames()]

        if isinstance(node.parent, nodes.ClassDef) and inferred_args:
            # first argument is special in these cases
            if node.type == "method":
                self.type_constraints.unify(inferred_args[0], ForwardRef(node.parent.name), node)
            elif node.type == "classmethod":
                self.type_constraints.unify(
                    inferred_args[0], Type[ForwardRef(node.parent.name)], node
                )

        # Get inferred return type
        if any(node.nodes_of_class(nodes.Return)):
            return_node = list(node.nodes_of_class(nodes.Return))[-1]
            if isinstance(return_node.inf_type, TypeFail):
                inferred_return = return_node.inf_type
            else:
                inferred_return = self.lookup_inf_type(node, "return")
        elif node.name == "__init__" and inferred_args:
            inferred_return = inferred_args[0]
        else:
            inferred_return = TypeInfo(type(None))

        # Update the environment storing the function's type.
        polymorphic_tvars = set()
        for arg in inferred_args + [inferred_return]:
            arg >> (lambda a: polymorphic_tvars.add(a.__name__) if isinstance(a, TypeVar) else None)

        # Create function signature
        func_type = create_Callable_TypeResult(
            failable_collect(inferred_args), inferred_return, polymorphic_tvars
        )

        # Check for optional arguments, create a Union of function signatures if necessary
        num_defaults = len(node.args.defaults)
        if num_defaults > 0 and not isinstance(func_type, TypeFail):
            for i in range(num_defaults):
                opt_args = inferred_args[: -1 - i]
                opt_func_type = create_Callable_TypeResult(
                    failable_collect(opt_args), inferred_return, polymorphic_tvars
                )
                func_type = func_type >> (
                    lambda f: opt_func_type >> (lambda opt_f: TypeInfo(Union[f, opt_f]))
                )

        # Final type signature unify
        func_name = self.lookup_inf_type(node.parent, node.name)
        result = self.type_constraints.unify(func_name, func_type, node)
        if isinstance(result, TypeFail):
            node.inf_type = result

    def visit_asyncfunctiondef(self, node: nodes.AsyncFunctionDef) -> None:
        self.visit_functiondef(node)

    def visit_lambda(self, node: nodes.Lambda) -> None:
        inferred_args = [self.lookup_inf_type(node, arg) for arg in node.argnames()]
        inferred_return = node.body.inf_type

        polymorphic_tvars = set()
        for arg in inferred_args + [inferred_return]:
            arg >> (lambda a: polymorphic_tvars.add(a.__name__) if isinstance(a, TypeVar) else None)

        node.inf_type = create_Callable_TypeResult(
            failable_collect(inferred_args), inferred_return, polymorphic_tvars
        )

    def visit_arguments(self, node: nodes.Arguments) -> None:
        node.inf_type = NoType()
        if any(annotation is not None for annotation in node.annotations):
            for i in range(len(node.annotations)):
                arg_tvar = self.lookup_typevar(node, node.args[i].name)

                if node.annotations[i] is not None:
                    ann_type = _ann_node_to_type(node.annotations[i])
                    result = self.type_constraints.unify(arg_tvar, ann_type, node)
                    if isinstance(result, TypeFail):
                        node.inf_type = result
                else:
                    self.type_constraints.unify(arg_tvar, Any, node)

    def visit_return(self, node: nodes.Return) -> None:
        return_tvar = self.lookup_typevar(node, "return")
        # TODO: Replace with isinstance() once proper TypeFail subclass is created for unbound indentifiers
        if return_tvar == TypeFail("Unbound identifier"):
            return_target = TypeFailReturn(node)
        else:
            return_target = return_tvar

        if node.value is not None and getattr(node.scope(), "returns", None) is not None:
            return_annotation = _ann_node_to_type(node.scope().returns)
            return_value = self.type_constraints.unify(node.value.inf_type, return_annotation, node)
        elif node.value is not None:
            return_value = node.value.inf_type
        else:
            return_value = TypeInfo(None)

        val_inf_type = self.type_constraints.unify(return_value, return_target, node)
        node.inf_type = val_inf_type if isinstance(val_inf_type, TypeFail) else NoType()

    def visit_classdef(self, node: nodes.ClassDef) -> None:
        node.inf_type = NoType()

        # Update type_store for this class.
        # TODO: include node.instance_attrs as well?
        for attr in node.locals:
            attr_inf_type = self.type_constraints.resolve(node.type_environment.lookup_in_env(attr))
            attr_inf_type >> (
                lambda a: self.type_store.methods[attr].append((a, node.locals[attr][0].type))
                if is_callable(a)
                else None
            )
            attr_inf_type >> (
                lambda a: self.type_store.classes[node.name][attr].append(
                    (a, node.locals[attr][0].type if is_callable(a) else "attribute")
                )
            )

    @accept_failable
    def get_attribute_class(self, t: type) -> Tuple[str, type, bool]:
        """Check for and return name and type of class represented by type t."""
        is_inst_expr = True

        # TypeVar; e.g., 'TypeVar('_T1')' corresponding to a function argument
        if isinstance(t, TypeVar):
            return t.__name__, t, None

        # Class type: e.g., 'Type[ForwardRef('A')]'
        if getattr(t, "__origin__", None) is type:
            class_type = t.__args__[0]
            is_inst_expr = False
        # Instance of class or builtin type; e.g., 'ForwardRef('A')' or 'int'
        else:
            class_type = t

        if isinstance(class_type, ForwardRef):
            class_name = class_type.__forward_arg__
        elif isinstance(class_type, _GenericAlias):
            class_name = class_type._name
        else:
            class_name = getattr(t, "__name__", None)

        # TODO: the condition below is too general
        if class_name is not None and class_name not in self.type_store.classes:
            class_name = class_name.lower()

        return class_name, class_type, is_inst_expr

    def visit_attribute(self, node: nodes.Attribute) -> None:
        expr_inf_type = self.type_constraints.resolve(node.expr.inf_type)
        result = self.get_attribute_class(expr_inf_type)
        if not isinstance(result, TypeFail):
            class_name, class_type, inst_expr = result

            if class_type == Any:
                node.inf_type = TypeInfo(Any)
            elif class_name in self.type_store.classes:
                attribute_type = None
                for par_class_type in self.type_store.classes[class_name]["__mro"]:
                    attribute_type = self.type_store.classes[par_class_type].get(node.attrname)
                    if attribute_type:
                        break
                if attribute_type is None:
                    class_tnode = self.type_constraints.get_tnode(class_type)
                    node.inf_type = TypeFailLookup(class_tnode, node, node.parent)
                else:
                    func_type, method_type = attribute_type[0]
                    if (
                        is_callable(func_type)
                        and method_type == "method"
                        and inst_expr
                        or method_type == "classmethod"
                    ):
                        # Replace polymorphic type variables with fresh type variables
                        fresh_func_type = self.type_constraints.fresh_callable(func_type, node)
                        self.type_constraints.unify(fresh_func_type.__args__[0], class_type)
                        # Create new Callable to avoid modifying elements of type store
                        new_func_type = create_Callable_TypeResult(
                            fresh_func_type.__args__[1:-1], fresh_func_type.__args__[-1]
                        )
                    else:
                        new_func_type = TypeInfo(func_type)
                    node.inf_type = new_func_type
            else:
                class_tnode = self.type_constraints.get_tnode(class_type)
                node.inf_type = TypeFailLookup(class_tnode, node, node.parent)
        else:
            node.inf_type = result

    def visit_module(self, node: nodes.Module) -> None:
        node.inf_type = NoType()


# Main function (useful for quick debugging)
def main(source: str) -> Tuple[nodes.Module, TypeInferer]:
    """Parse a string representing source text, and perform a typecheck.

    Return the astroid Module node (with the type_constraints attribute set
    on all nodes in the tree) and TypeInferer object.
    """
    module = astroid.parse(source)
    type_inferer = TypeInferer()
    type_inferer.environment_transformer().visit(module)
    type_inferer.type_inference_transformer().visit(module)
    return module, type_inferer



import sys

import pytest
from astroid import nodes
from hypothesis import HealthCheck, assume, given, settings

from .. import custom_hypothesis_support as cs

settings.load_profile("pyta")


@given(cs.subscript_node())
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_index(node):
    if sys.version_info >= (3, 9):
        pytest.skip("Index node is deprecated in Python 3.9")
    module, _ = cs._parse_text(node)
    for index_node in module.nodes_of_class(nodes.Index):
        assert index_node.inf_type.getValue() == index_node.value.inf_type.getValue()


@given(cs.expr_node())
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_expr(expr):
    module, _ = cs._parse_text(expr)
    for expr_node in module.nodes_of_class(nodes.Expr):
        assert expr_node.inf_type.getValue() == expr_node.value.inf_type.getValue()



"""checker for variable shadowing in a comprehension.
"""
from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker


class ShadowingInComprehensionChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = "shadowing_in_comprehension"
    msgs = {
        "E9988": (
            "Comprehension variable '%s' shadows a variable in an outer scope",
            "shadowing-in-comprehension",
            "Used when there is shadowing inside a comprehension",
        )
    }

    # this is important so that your checker is executed before others
    priority = -1

    @check_messages("shadowing-in-comprehension")
    def visit_comprehension(self, node: nodes.Comprehension):
        if isinstance(node.target, nodes.Tuple):
            for target in node.target.elts:
                if target.name in node.parent.frame().locals and target.name != "_":
                    args = target.name
                    self.add_message("shadowing-in-comprehension", node=target, args=args)
        else:  # isinstance(node.target, nodes.AssignName)
            if node.target.name in node.parent.frame().locals and node.target.name != "_":
                args = node.target.name
                self.add_message("shadowing-in-comprehension", node=node.target, args=args)


def register(linter):
    linter.register_checker(ShadowingInComprehensionChecker(linter))



from typing import ForwardRef

from astroid import nodes

from python_ta.transforms.type_inference_visitor import TypeFail

from .. import custom_hypothesis_support as cs


def test_class_with_init():
    program = """
    class Foo:

        def __init__(self):
            self.a = 5

    foo = Foo()
    """
    ast_mod, ti = cs._parse_text(program)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert isinstance(call_node.inf_type.getValue(), ForwardRef)
        assert call_node.inf_type.getValue() == ForwardRef("Foo")


def test_class_without_init():
    program = """
    class Foo:
        def fee(self):
            return 1

    foo = Foo()
    """
    ast_mod, ti = cs._parse_text(program)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert isinstance(call_node.inf_type.getValue(), ForwardRef)
        assert call_node.inf_type.getValue() == ForwardRef("Foo")


def test_wrong_number_init():
    program = """
    class Foo:
        def __init__(self, x):
            self.a = x

    foo = Foo()
    """
    ast_mod, ti = cs._parse_text(program, True)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert isinstance(call_node.inf_type, TypeFail)


def test_class_defined_later():
    program = """
    class A:
        def __init__(self):
            self.attr = B()

    class B:
        pass
    """
    ast_mod, ti = cs._parse_text(program, True)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        assert not isinstance(call_node.inf_type, TypeFail)


def test_builtin_overloaded_initializers():
    program = """
    range1 = range(10)
    range2 = range(1, 10, 1)
    """
    ast_mod, ti = cs._parse_text(program, True)
    for assgn_node in ast_mod.nodes_of_class(nodes.AssignName):
        if assgn_node.name == "range1":
            range1 = ti.lookup_typevar(assgn_node, assgn_node.name)
            assert ti.type_constraints.resolve(range1).getValue() == range
        if assgn_node.name == "range2":
            range2 = ti.lookup_typevar(assgn_node, assgn_node.name)
            assert ti.type_constraints.resolve(range2).getValue() == range



from typing import *

from astroid import nodes
from pytest import skip

from python_ta.typecheck.base import TypeFail

from .. import custom_hypothesis_support as cs


def test_instance_dot_method():
    program = """
        class A:
            def foo(self, x):
                return x + 1

        A().foo(0)
        """
    module, _ = cs._parse_text(program, reset=True)
    for attribute_node in module.nodes_of_class(nodes.Attribute):
        assert str(attribute_node.inf_type.getValue()) == "typing.Callable[[int], int]"


def test_instance_dot_classmethod():
    program = """
        class A:
            @classmethod
            def foo(cls, x):
                return x + 1

        A().foo(0)
        """
    module, _ = cs._parse_text(program, reset=True)
    for attribute_node in module.nodes_of_class(nodes.Attribute):
        assert str(attribute_node.inf_type.getValue()) == "typing.Callable[[int], int]"


def test_instance_dot_staticmethod():
    program = """
        class A:
            @staticmethod
            def foo(x):
                return x + 1

        A().foo(0)
        """
    module, _ = cs._parse_text(program, reset=True)
    for attribute_node in module.nodes_of_class(nodes.Attribute):
        assert str(attribute_node.inf_type.getValue()) == "typing.Callable[[int], int]"


def test_class_dot_method():
    program = """
        class A:
            def foo(self, x):
                return x + 1

        A.foo(A(), 0)
        """
    module, _ = cs._parse_text(program, reset=True)
    for attribute_node in module.nodes_of_class(nodes.Attribute):
        assert (
            str(attribute_node.inf_type.getValue())
            == "typing.Callable[[ForwardRef('A'), int], int]"
        )


def test_class_dot_classmethod():
    program = """
        class A:
            @classmethod
            def foo(cls, x):
                return x + 1

        A.foo(0)
        """
    module, _ = cs._parse_text(program, reset=True)
    for attribute_node in module.nodes_of_class(nodes.Attribute):
        assert str(attribute_node.inf_type.getValue()) == "typing.Callable[[int], int]"


def test_class_dot_staticmethod():
    program = """
        class A:
            @staticmethod
            def foo(x):
                return x + 1

        A.foo(0)
        """
    module, _ = cs._parse_text(program, reset=True)
    for attribute_node in module.nodes_of_class(nodes.Attribute):
        assert str(attribute_node.inf_type.getValue()) == "typing.Callable[[int], int]"


def test_attribute_self_bind():
    """Make sure auto-binding of self persists"""
    program = """
        x = []
        f = x.append
        f(4)
        """
    module, ti = cs._parse_text(program, reset=True)
    x = [ti.lookup_typevar(node, node.name) for node in module.nodes_of_class(nodes.AssignName)][0]
    assert str(ti.type_constraints.resolve(x).getValue()) == "typing.List[int]"


def test_subscript_attribute():
    program = """
        class A:
            def __init__(self):
                self.lst = []
                self.lst[0] = 1

        a = A()
        x = a.lst[0]
        """
    module, ti = cs._parse_text(program, reset=True)
    for assgn_node in module.nodes_of_class(nodes.AssignName):
        if assgn_node.name == "x":
            x = ti.lookup_typevar(assgn_node, assgn_node.name)
            assert ti.type_constraints.resolve(x).getValue() == int


def test_obj_list_attribute():
    program = """
        class A:
            def __init__(self):
                self.name = 'abc'

        lst = [A(), A()]
        x = lst[0].name
        """
    module, ti = cs._parse_text(program, reset=True)
    for assgn_node in module.nodes_of_class(nodes.AssignName):
        if assgn_node.name == "x":
            x = ti.lookup_typevar(assgn_node, assgn_node.name)
            assert ti.type_constraints.resolve(x).getValue() == str


def test_unknown_class_attribute():
    program = """
        def foo(a, x):
            y = x.name
            y = 3
            z = x.name
        """
    module, ti = cs._parse_text(program, reset=True)
    for assgn_node in module.nodes_of_class(nodes.AssignName):
        if assgn_node.name == "z":
            z = ti.lookup_typevar(assgn_node, assgn_node.name)
            assert ti.type_constraints.resolve(z).getValue() == int


def test_unknown_class_attribute2():
    program = """
        def foo(x):
            return x.name + 2

        def bar(y):
            z = foo(y)
        """
    module, ti = cs._parse_text(program, reset=True)
    for binop_node in module.nodes_of_class(nodes.BinOp):
        assert binop_node.inf_type.getValue() == int
    for assgn_node in module.nodes_of_class(nodes.AssignName):
        if assgn_node.name == "z":
            z = ti.lookup_typevar(assgn_node, assgn_node.name)
            assert ti.type_constraints.resolve(z).getValue() == int


def test_unknown_class_subscript_attribute():
    skip("Currently, the inferred type of the z variable is str")
    program = """
        def foo(x):
            return x.name[0]

        def bar(y):
            z = foo(y)
        """
    module, ti = cs._parse_text(program, reset=True)
    for assgn_node in module.nodes_of_class(nodes.AssignName):
        if assgn_node.name == "z":
            z = ti.lookup_typevar(assgn_node, assgn_node.name)
            assert isinstance(ti.type_constraints.resolve(z).getValue(), TypeVar)


def test_unknown_class_subscript_attribute2():
    program = """
        def foo(x):
            x.name[0] = 2
            return x.name[1]

        def bar(y):
            z = foo(y)
        """
    module, ti = cs._parse_text(program, reset=True)
    for assgn_node in module.nodes_of_class(nodes.AssignName):
        if assgn_node.name == "z":
            z = ti.lookup_typevar(assgn_node, assgn_node.name)
            assert ti.type_constraints.resolve(z).getValue() == int


def test_invalid_builtin_attribute():
    program = """
        x = 3
        y = x.name
        """
    module, ti = cs._parse_text(program, reset=True)
    assgn_node = list(module.nodes_of_class(nodes.Assign))[1]
    assert isinstance(assgn_node.inf_type, TypeFail)


def test_attribute_unification_fail():
    skip("This case is not supported yet")
    program = """
        class A:
            def __init__(self):
                self.name = 'abc'

        def f(x):
            x.name = 3

        a = A()
        f(a)
        f(A)
        """
    module, ti = cs._parse_text(program, reset=True)
    call_node1, call_node2 = module.nodes_of_class(nodes.Call)[1:]
    assert isinstance(call_node.inf_type1, TypeFail)
    assert isinstance(call_node.inf_type2, TypeFail)



"""Monkeypatch pylint behaviour.
"""
from .checkers import patch_checkers
from .error_messages import patch_error_messages
from .messages import patch_linter_transform, patch_messages
from .transforms import patch_ast_transforms


def patch_all():
    """Execute all patches defined in this module."""
    patch_checkers()
    patch_ast_transforms()
    patch_messages()
    patch_linter_transform()
    patch_error_messages()



from typing import Any

from astroid import nodes

from python_ta.typecheck.base import TypeFail

from .. import custom_hypothesis_support as cs


def test_imported_module_attribute():
    src = """
    import mod

    y = mod.x
    z = undefined_mod.x
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for assgn_node in ast_mod.nodes_of_class(nodes.Assign):
        if assgn_node.targets[0].name == "y":
            y = ti.lookup_typevar(assgn_node, "y")
            assert ti.type_constraints.resolve(y).getValue() == Any
        if assgn_node.targets[0].name == "z":
            assert isinstance(assgn_node.inf_type, TypeFail)


def test_imported_module_attribute_persists_any():
    src = """
    import mod

    y = mod.x
    y = 3
    z = mod.x
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for assgn_node in ast_mod.nodes_of_class(nodes.Assign):
        if assgn_node.targets[0].name == "z":
            z = ti.lookup_typevar(assgn_node, "z")
            assert ti.type_constraints.resolve(z).getValue() == Any


def test_from_import_variable():
    src = """
    from mod import x

    y = x
    z = w
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for assgn_node in ast_mod.nodes_of_class(nodes.Assign):
        if assgn_node.targets[0].name == "y":
            y = ti.lookup_typevar(assgn_node, "y")
            assert ti.type_constraints.resolve(y).getValue() == Any
        if assgn_node.targets[0].name == "z":
            assert isinstance(assgn_node.inf_type, TypeFail)


def test_from_import_attribute():
    src = """
    from mod import cls

    y = cls.x1
    z = undefined_cls.x2
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for assgn_node in ast_mod.nodes_of_class(nodes.Assign):
        if assgn_node.targets[0].name == "y":
            y = ti.lookup_typevar(assgn_node, "y")
            assert ti.type_constraints.resolve(y).getValue() == Any
        if assgn_node.targets[0].name == "z":
            assert isinstance(assgn_node.inf_type, TypeFail)


def test_imported_module_function():
    src = """
    import mod

    x = mod.func1(0, 1)
    y = mod.func1(0, 1) + 1
    undefined_mod.func2(0, 1)
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        if call_node.func.attrname == "func1":
            assert call_node.inf_type.getValue() == Any
        if call_node.func.attrname == "func2":
            assert isinstance(call_node.inf_type, TypeFail)
    for assgn_node in ast_mod.nodes_of_class(nodes.Assign):
        if assgn_node.targets[0].name == "x":
            x = ti.lookup_typevar(assgn_node, "x")
            assert ti.type_constraints.resolve(x).getValue() == Any
        if assgn_node.targets[0].name == "y":
            y = ti.lookup_typevar(assgn_node, "y")
            assert ti.type_constraints.resolve(y).getValue() == int


def test_from_import_function():
    src = """
    from mod import func

    x = func(0, 1)
    y = func(0, 1) + 1
    undefined_func(0, 1)
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    for call_node in ast_mod.nodes_of_class(nodes.Call):
        if call_node.func.name == "func1":
            assert call_node.inf_type.getValue() == Any
        if call_node.func.name == "func2":
            assert isinstance(call_node.inf_type, TypeFail)
    for assgn_node in ast_mod.nodes_of_class(nodes.Assign):
        if assgn_node.targets[0].name == "x":
            x = ti.lookup_typevar(assgn_node, "x")
            assert ti.type_constraints.resolve(x).getValue() == Any
        if assgn_node.targets[0].name == "y":
            y = ti.lookup_typevar(assgn_node, "y")
            assert ti.type_constraints.resolve(y).getValue() == int



"""Checker for target of for loop in subscript form.
"""
from typing import List, Union

from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.checkers.utils import check_messages
from pylint.interfaces import IAstroidChecker


class InvalidForTargetChecker(BaseChecker):
    __implements__ = IAstroidChecker
    # name is the same as file name but without _checker part
    name = "invalid_for_target"
    # use dashes for connecting words in message symbol
    msgs = {
        "E9984": (
            'For loop variable "%s" should not be a part of a larger object.',
            "invalid-for-target",
            "Used when you have a loop variable in a for loop "
            "that is in subscript or object attribute form",
        )
    }
    # this is important so that your checker is executed before others
    priority = -1

    INVALID_TARGETS = (nodes.Subscript, nodes.AssignAttr)

    @check_messages("invalid-for-target")
    def visit_for(self, node: nodes.For) -> None:
        invalid_for_targets = node.target.nodes_of_class(self.INVALID_TARGETS)
        for target in invalid_for_targets:
            self.add_message("invalid-for-target", node=target, args=target.as_string())


def register(linter):
    linter.register_checker(InvalidForTargetChecker(linter))



"""checker for global variables
"""
import re

from astroid import nodes
from pylint.checkers import BaseChecker
from pylint.checkers.base import UpperCaseStyle
from pylint.interfaces import IAstroidChecker


class GlobalVariablesChecker(BaseChecker):

    __implements__ = IAstroidChecker

    name = "global_variables"
    msgs = {
        "E9997": (
            "Global variables must be constants in CSC108/CSC148: " "%s",
            "forbidden-global-variables",
            "",
        )
    }

    # this is important so that your checker is executed before others
    priority = -1

    def __init__(self, linter=None):
        super().__init__(linter)
        self.import_names = []

    def visit_global(self, node):
        args = "the keyword 'global' is used on line {}".format(node.lineno)
        self.add_message("forbidden-global-variables", node=node, args=args)

    def visit_assignname(self, node):
        """Allow global constant variables (uppercase), but issue messages for
        all other globals.
        """
        self._inspect_vars(node)

    def visit_name(self, node):
        """Allow global constant variables (uppercase), but issue messages for
        all other globals.
        """
        self._inspect_vars(node)

    def visit_import(self, node):
        """Save the names of imports, to prevent mistaking for global vars."""
        self._store_name_or_alias(node)

    def visit_importfrom(self, node):
        """Save the names of imports, to prevent mistaking for global vars."""
        self._store_name_or_alias(node)

    def _store_name_or_alias(self, node):
        for name_tuple in node.names:
            if name_tuple[1] is not None:
                self.import_names.append(name_tuple[1])  # aliased names
            else:
                self.import_names.append(name_tuple[0])  # no alias

    def _inspect_vars(self, node):
        """Allows constant, global variables (i.e. uppercase), but issue
        messages for all other global variables.
        """
        if hasattr(node, "name") and node.name in self.import_names:
            return
        if isinstance(node.frame(), nodes.Module) and not is_in_main(node):
            node_list = _get_child_disallowed_global_var_nodes(node)
            for node in node_list:
                if isinstance(node, nodes.AssignName):
                    args = "a global variable '{}' is assigned to on line {}".format(
                        node.name, node.lineno
                    )
                else:
                    args = "a global variable '{}' is used on line {}".format(
                        node.name, node.lineno
                    )
                self.add_message("forbidden-global-variables", node=node, args=args)


def _get_child_disallowed_global_var_nodes(node):
    """Return a list of all top-level Name or AssignName nodes for a given
    global, non-constant variable.

    TODO: use the configured NamingStyle instead of hard-coded SnakeCaseStyle
    for the CONST_NAME_RGX value.
    """
    node_list = []
    if (
        (
            isinstance(node, (nodes.AssignName, nodes.Name))
            and not isinstance(node.parent, nodes.Call)
        )
        and not re.match(UpperCaseStyle.CONST_NAME_RGX, node.name)
        and node.scope() is node.root()
    ):
        return [node]

    for child_node in node.get_children():
        node_list += _get_child_disallowed_global_var_nodes(child_node)
    return node_list


def is_in_main(node):
    if not hasattr(node, "parent"):
        return False

    parent = node.parent
    try:
        if (
            isinstance(parent, nodes.If)
            and parent.test.left.name == "__name__"
            and parent.test.ops[0][1].value == "__main__"
        ):
            return True
        else:
            return is_in_main(parent)
    except (AttributeError, IndexError) as e:
        return is_in_main(parent)


def register(linter):
    """required method to auto register this checker"""
    linter.register_checker(GlobalVariablesChecker(linter))



from typing import Any, Dict, List, Set, Tuple, Union, _GenericAlias

import hypothesis.strategies as hs
from astroid import nodes
from hypothesis import HealthCheck, given, settings
from pytest import skip

from python_ta.typecheck.base import (
    NoType,
    TypeFail,
    TypeFailAnnotationInvalid,
    TypeFailUnify,
    _gorg,
    _node_to_type,
)

from .. import custom_hypothesis_support as cs
from ..custom_hypothesis_support import lookup_type

settings.load_profile("pyta")


def test_annassign_concrete():
    """Test whether types are being properly set for an AnnAssign node."""
    program = (
        f"class Student:\n"
        f"    name: str\n"
        f"    age: int\n"
        f"    status: bool\n"
        f"    def __init__(self):\n"
        f"        pass\n"
        f""
    )
    module, inferer = cs._parse_text(program)
    for node in module.nodes_of_class(nodes.AnnAssign):
        variable_type = lookup_type(inferer, node, node.target.name)
        annotated_type = _node_to_type(node.annotation.name)
        assert variable_type == annotated_type


@given(hs.dictionaries(cs.valid_identifier(), cs.builtin_type, min_size=2))
@settings(suppress_health_check=[HealthCheck.too_slow])
def test_annassign(variables_annotations_dict):
    """Test whether types are being properly set for an AnnAssign node."""
    program = f"class Student:\n"
    for variable in variables_annotations_dict:
        program += f"    {variable}: {variables_annotations_dict[variable].__name__}\n"
    program += f"    def __init__(self):\n" f"        pass\n"
    module, inferer = cs._parse_text(program)
    for node in module.nodes_of_class(nodes.AnnAssign):
        variable_type = lookup_type(inferer, node, node.target.name)
        annotated_type = variables_annotations_dict[node.target.name]
        if isinstance(variable_type, _GenericAlias):
            assert _gorg(variable_type) == annotated_type
        else:
            assert variable_type == annotated_type


def test_annassign_subscript_list():
    program = """
    lst: List
    """
    module, inferer = cs._parse_text(program)
    ann_node = next(module.nodes_of_class(nodes.AnnAssign))
    variable_type = lookup_type(inferer, ann_node, ann_node.target.name)
    t = inferer.type_constraints.resolve(variable_type)
    assert t.getValue() == List[Any]


def test_annassign_subscript_list_int():
    program = """
    lst: List[int]

    lst = [1, 2, 3]
    """
    module, inferer = cs._parse_text(program)
    ann_node = next(module.nodes_of_class(nodes.AnnAssign))
    variable_type = lookup_type(inferer, ann_node, ann_node.target.name)
    assert variable_type == List[int]

    assign_node = next(module.nodes_of_class(nodes.Assign))
    assign_type = lookup_type(inferer, assign_node, assign_node.targets[0].name)
    assert assign_type == List[int]


def test_annassign_subscript_list_int_wrong():
    program = """
    lst: List[int]

    lst = ['Hello', 'Goodbye']
    """
    module, inferer = cs._parse_text(program)
    ann_node = next(module.nodes_of_class(nodes.AnnAssign))
    variable_type = lookup_type(inferer, ann_node, ann_node.target.name)
    assert variable_type == List[int]

    assign_node = next(module.nodes_of_class(nodes.Assign))
    assert isinstance(assign_node.inf_type, TypeFailUnify)


def test_annassign_subscript_set():
    program = """
    s: Set
    """
    module, inferer = cs._parse_text(program)
    ann_node = next(module.nodes_of_class(nodes.AnnAssign))
    variable_type = lookup_type(inferer, ann_node, ann_node.target.name)
    assert variable_type == Set[Any]


def test_annassign_subscript_set_int():
    program = """
    s: Set[int]
    """
    module, inferer = cs._parse_text(program)
    ann_node = next(module.nodes_of_class(nodes.AnnAssign))
    variable_type = lookup_type(inferer, ann_node, ann_node.target.name)
    assert variable_type == Set[int]


def test_annassign_subscript_dict():
    program = """
    d: Dict
    """
    module, inferer = cs._parse_text(program)
    ann_node = next(module.nodes_of_class(nodes.AnnAssign))
    variable_type = lookup_type(inferer, ann_node, ann_node.target.name)
    assert variable_type == Dict[Any, Any]


def test_annassign_subscript_dict_int_str():
    program = """
    d: Dict[int, str]
    """
    module, inferer = cs._parse_text(program)
    ann_node = next(module.nodes_of_class(nodes.AnnAssign))
    variable_type = lookup_type(inferer, ann_node, ann_node.target.name)
    assert variable_type == Dict[int, str]


def test_annassign_subscript_tuple():
    program = """
    t: Tuple
    """
    module, inferer = cs._parse_text(program)
    ann_node = next(module.nodes_of_class(nodes.AnnAssign))
    variable_type = lookup_type(inferer, ann_node, ann_node.target.name)
    assert variable_type == Tuple[Any]


def test_annassign_subscript_tuple_int():
    program = """
    t: Tuple[int, int]
    """
    module, inferer = cs._parse_text(program)
    ann_node = next(module.nodes_of_class(nodes.AnnAssign))
    variable_type = lookup_type(inferer, ann_node, ann_node.target.name)
    assert variable_type == Tuple[int, int]


def test_annassign_subscript_tuple_multi_param():
    program = """
    t: Tuple

    t = (1, 'Hello')
    """
    skip("Requires support for multi-parameter Tuple annotations")
    module, inferer = cs._parse_text(program)
    ann_node = next(module.nodes_of_class(nodes.AnnAssign))
    variable_type = lookup_type(inferer, ann_node, ann_node.target.name)
    assert variable_type == Tuple[int, int]


def test_annassign_subscript_multi_list():
    program = """
    l1: List
    l2: List

    l1 = [1, 2, 3]
    l2 = ['Hello', 'Goodbye']
    """
    module, inferer = cs._parse_text(program)

    for ann_node in module.nodes_of_class(nodes.AnnAssign):
        variable_type = lookup_type(inferer, ann_node, ann_node.target.name)
        assert variable_type == List[Any]

    assign_nodes = list(module.nodes_of_class(nodes.Assign))

    assign_node_1 = assign_nodes[0]
    assign_type_1 = lookup_type(inferer, assign_node_1, assign_node_1.targets[0].name)
    assert assign_type_1 == List[Any]

    assign_node_2 = assign_nodes[1]
    assign_type_2 = lookup_type(inferer, assign_node_2, assign_node_2.targets[0].name)
    assert assign_type_2 == List[Any]


def test_annassign_and_assign():
    src = """
    x: List[int] = [1, 2, 3]
    """
    module, inferer = cs._parse_text(src, reset=True)
    x = [
        inferer.lookup_typevar(node, node.name) for node in module.nodes_of_class(nodes.AssignName)
    ][0]
    for ann_node in module.nodes_of_class(nodes.AnnAssign):
        assert ann_node.inf_type == NoType()
    assert inferer.type_constraints.resolve(x).getValue() == List[int]


def test_invalid_annassign_and_assign():
    src = """
    x: List[str] = [1, 2, 3]
    """
    module, inferer = cs._parse_text(src, reset=True)
    for ann_node in module.nodes_of_class(nodes.AnnAssign):
        assert isinstance(ann_node.inf_type, TypeFailUnify)


def test_annotation_not_type():
    src = """
    x: [str, int]
    """
    module, inferer = cs._parse_text(src, reset=True)
    for ann_node in module.nodes_of_class(nodes.AnnAssign):
        assert isinstance(ann_node.inf_type, TypeFailAnnotationInvalid)


def test_annotation_forward_ref():
    src = """
    x: 'SomeClass'
    """
    module, inferer = cs._parse_text(src, reset=True)
    for ann_node in module.nodes_of_class(nodes.AnnAssign):
        assert not isinstance(ann_node.inf_type, TypeFail)


def test_param_annotation_not_type():
    src = """
    def f(x: [str, int]) -> None:
        return x
    """
    module, inferer = cs._parse_text(src, reset=True)
    for arg_node in module.nodes_of_class(nodes.Arguments):
        assert isinstance(arg_node.inf_type, TypeFailAnnotationInvalid)


def test_annotation_forward_ref_space():
    src = """
    x: 'Some Class'
    """
    module, inferer = cs._parse_text(src, reset=True)
    for ann_node in module.nodes_of_class(nodes.AnnAssign):
        assert isinstance(ann_node.inf_type, TypeFailAnnotationInvalid)


def test_annotation_union_list():
    src = """
    x: Union[List, int]
    """
    module, inferer = cs._parse_text(src, reset=True)
    for ann_node in module.nodes_of_class(nodes.AnnAssign):
        assert not isinstance(ann_node.inf_type, TypeFail)
    x_type = lookup_type(inferer, module, "x")
    assert x_type == Union[List[Any], int]



from typing import *
from typing import ForwardRef, _GenericAlias

import pytest
from astroid import nodes

from python_ta.typecheck.base import TypeConstraints, TypeFail, _TNode
from sample_usage.draw_tnodes import gen_graph_from_nodes

from .. import custom_hypothesis_support as cs


def tc_to_disjoint(tc: TypeConstraints) -> List[Set[Union[type, str]]]:
    tnode_list = tc._nodes.copy()
    disjoint_set = []
    while tnode_list:
        cur_tnode = tnode_list[0]
        new_set = set()
        open_nodes = [cur_tnode]
        visited_nodes = []
        while open_nodes:
            cur_open_node = open_nodes[0]
            if isinstance(cur_open_node.type, TypeVar) or isinstance(
                cur_open_node.type, _GenericAlias
            ):
                new_set.add(str(cur_open_node.type))
            else:
                new_set.add(cur_open_node.type)
            for e in cur_open_node.adj_list:
                if e[0] not in visited_nodes and e[0] not in open_nodes:
                    open_nodes.append(e[0])
            visited_nodes.append(cur_open_node)
            if cur_open_node in tnode_list:
                tnode_list.remove(cur_open_node)
            open_nodes.remove(cur_open_node)
        disjoint_set.append(new_set)
    return disjoint_set


def compare_list_sets(s1: List[Set[Union[type, str]]], s2: List[Set[Union[type, str]]]):
    list_sets_1 = [{str(t) for t in s} for s in s1]
    list_sets_2 = [{str(t) for t in s} for s in s2]
    assert len(list_sets_1) == len(list_sets_2)
    while list_sets_1:
        set1 = list_sets_1[0]
        found_set = None
        for set2 in list_sets_2:
            if set1 == set2:
                found_set = set2
        if found_set:
            list_sets_1.remove(set1)
            list_sets_2.remove(found_set)
        else:
            assert False
    assert True


def test_one_var():
    src = """
    a = 1
    """
    ast_mod, ti = cs._parse_text(src)
    actual_set = tc_to_disjoint(ti.type_constraints)
    expected_set = [{int, "~_TV0"}]
    compare_list_sets(actual_set, expected_set)


def test_typefail():
    src = """
    a = 1
    b = "hello"
    a = b
    """
    ast_mod, ti = cs._parse_text(src)
    actual_set = tc_to_disjoint(ti.type_constraints)
    expected_set = [{int, "~_TV0"}, {str, "~_TV1"}]
    compare_list_sets(actual_set, expected_set)


def test_multi_var_int():
    src = """
    a = 1
    b = 2
    c = 3
    d = 4
    """
    ast_mod, ti = cs._parse_text(src)
    actual_set = tc_to_disjoint(ti.type_constraints)
    expected_set = [{int, "~_TV0", "~_TV1", "~_TV2", "~_TV3"}]
    compare_list_sets(actual_set, expected_set)


def test_multi_var():
    src = """
    a = 1
    b = "Hello"
    c = False
    d = 4.0
    """
    ast_mod, ti = cs._parse_text(src)
    actual_set = tc_to_disjoint(ti.type_constraints)
    expected_set = [{int, "~_TV0"}, {"~_TV1", str}, {"~_TV2", bool}, {"~_TV3", float}]
    compare_list_sets(actual_set, expected_set)


def test_var_chain():
    src = """
    a = 1
    b = a
    c = b
    d = c
    e = d
    """
    ast_mod, ti = cs._parse_text(src)
    actual_set = tc_to_disjoint(ti.type_constraints)
    expected_set = [{int, "~_TV0", "~_TV1", "~_TV2", "~_TV3", "~_TV4"}]
    compare_list_sets(actual_set, expected_set)


def test_list_int():
    src = """
    l = [1,2,3]
    """
    ast_mod, ti = cs._parse_text(src)
    actual_set = tc_to_disjoint(ti.type_constraints)
    expected_set = [{"~_TV0", "typing.List[int]"}, {int}]
    compare_list_sets(actual_set, expected_set)


def test_any_list():
    src = """
    l = [1, "string", False]
    """
    ast_mod, ti = cs._parse_text(src)
    actual_set = tc_to_disjoint(ti.type_constraints)
    expected_set = [{"~_TV0", "typing.List[typing.Any]"}, {int}, {str}]
    compare_list_sets(actual_set, expected_set)


def test_tuple_int():
    src = """
    l = (1, 2)
    """
    ast_mod, ti = cs._parse_text(src)
    actual_set = tc_to_disjoint(ti.type_constraints)
    expected_set = [{"~_TV0", "typing.Tuple[int, int]"}]
    compare_list_sets(actual_set, expected_set)


def test_tuple_int_str():
    src = """
    l = (1, "Hello")
    """
    ast_mod, ti = cs._parse_text(src)
    actual_set = tc_to_disjoint(ti.type_constraints)
    expected_set = [{"~_TV0", "typing.Tuple[int, str]"}]
    compare_list_sets(actual_set, expected_set)


def test_elt():
    src = """
    l = [1,2,3]
    e = l[0]
    """
    ast_mod, ti = cs._parse_text(src)
    actual_set = tc_to_disjoint(ti.type_constraints)
    expected_set = [{"~_TV0", "typing.List[~_TV2]", "typing.List[int]"}, {"~_TV1", int, "~_TV2"}]
    compare_list_sets(actual_set, expected_set)


tc = TypeConstraints()


def test_forward_ref(draw=False):
    tc.reset()
    t0 = tc.fresh_tvar()
    assert isinstance(tc.unify(ForwardRef("A"), ForwardRef("B")), TypeFail)
    assert tc.unify(ForwardRef("A"), ForwardRef("A")).getValue() == ForwardRef("A")
    assert tc.unify(t0, ForwardRef("A")).getValue() == ForwardRef("A")
    actual_set = tc_to_disjoint(tc)
    expected_set = [{"~_TV0", ForwardRef("A")}, {ForwardRef("B")}]
    compare_list_sets(actual_set, expected_set)
    if draw:
        gen_graph_from_nodes(tc._nodes)


def test_polymorphic_callable(draw=False):
    tc.reset()
    t0 = tc.fresh_tvar()
    t1 = tc.fresh_tvar()
    t2 = tc.fresh_tvar()
    tc.unify(t1, Callable[[int], int])
    tc.unify(t2, Callable[[t0], t0])
    tc.unify(t1, t2)
    actual_set = tc_to_disjoint(tc)
    expected_set = [
        {"~_TV0", int},
        {"~_TV1", "~_TV2", "typing.Callable[[int], int]", "typing.Callable[[~_TV0], ~_TV0]"},
    ]
    compare_list_sets(actual_set, expected_set)
    if draw:
        gen_graph_from_nodes(tc._nodes)


def test_polymorphic_callable2(draw=False):
    tc.reset()
    t0 = tc.fresh_tvar()
    t1 = tc.fresh_tvar()
    t2 = tc.fresh_tvar()
    tc.unify(t1, Callable[[t0], int])
    tc.unify(t2, Callable[[int], t0])
    tc.unify(t1, t2)
    actual_set = tc_to_disjoint(tc)
    expected_set = [
        {"~_TV0", int},
        {"~_TV1", "~_TV2", "typing.Callable[[~_TV0], int]", "typing.Callable[[int], ~_TV0]"},
    ]
    compare_list_sets(actual_set, expected_set)
    if draw:
        gen_graph_from_nodes(tc._nodes)


def test_polymorphic_callable3(draw=False):
    tc.reset()
    t0 = tc.fresh_tvar()
    t1 = tc.fresh_tvar()
    t2 = tc.fresh_tvar()
    tc.unify(t1, Callable[[t0], t0])
    tc.unify(t2, t1)
    tc.unify(t2, Callable[[int], int])
    actual_set = tc_to_disjoint(tc)
    expected_set = [
        {"~_TV0", int},
        {"~_TV1", "~_TV2", "typing.Callable[[int], int]", "typing.Callable[[~_TV0], ~_TV0]"},
    ]
    compare_list_sets(actual_set, expected_set)
    if draw:
        gen_graph_from_nodes(tc._nodes)


def test_polymorphic_callable4(draw=False):
    tc.reset()
    t0 = tc.fresh_tvar()
    t1 = tc.fresh_tvar()
    t2 = tc.fresh_tvar()
    tc.unify(t1, Callable[[t0], t0])
    tc.unify(t2, t1)
    assert isinstance(tc.unify(t2, Callable[[int], str]), TypeFail)
    actual_set = tc_to_disjoint(tc)
    expected_set = [
        {"~_TV0", int},
        {"~_TV1", "~_TV2", "typing.Callable[[~_TV0], ~_TV0]"},
        {"typing.Callable[[int], str]"},
        {str},
    ]
    compare_list_sets(actual_set, expected_set)
    if draw:
        gen_graph_from_nodes(tc._nodes)


def test_polymorphic_callable5(draw=False):
    tc.reset()
    t0 = tc.fresh_tvar()
    tc.unify(
        Callable[[Callable[[int, t0], int], int], t0],
        Callable[[Callable[[int, int], int], int], t0],
    )
    actual_set = tc_to_disjoint(tc)
    expected_set = [
        {"~_TV0", int},
        {"typing.Callable[[int, ~_TV0], int]", "typing.Callable[[int, int], int]"},
        {
            "typing.Callable[[typing.Callable[[int, ~_TV0], int], int], ~_TV0]",
            "typing.Callable[[typing.Callable[[int, int], int], int], ~_TV0]",
        },
    ]
    compare_list_sets(actual_set, expected_set)
    if draw:
        gen_graph_from_nodes(tc._nodes)


def test_can_unify_callable(draw=False):
    tc.reset()
    t0 = tc.fresh_tvar()
    assert not tc.can_unify(Callable[[t0, t0], int], Callable[[str, int], int])
    # make sure tc is unchanged
    actual_set = tc_to_disjoint(tc)
    expected_set = [{"~_TV0"}]
    compare_list_sets(actual_set, expected_set)
    if draw:
        gen_graph_from_nodes(tc._nodes)


# Inheritance


def test_userdefn_inheritance_simple(draw=False):
    src = """
    class A:
        pass

    class B:
        pass

    class C(A, B):
        pass

    a = A()
    b = B()
    c = C()
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tc = ti.type_constraints
    a, b, c = [
        ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)
    ]

    assert isinstance(tc.unify(a, b), TypeFail)
    assert tc.unify(c, a).getValue() == ForwardRef("C")
    assert isinstance(tc.unify(a, c), TypeFail)  # note that order matters!
    assert tc.unify(c, b).getValue() == ForwardRef("C")
    assert isinstance(tc.unify(b, c), TypeFail)

    actual_set = tc_to_disjoint(tc)
    expected_set = [
        {"~_TV0", Type[ForwardRef("A")]},
        {"~_TV1", Type[ForwardRef("B")]},
        {"~_TV2", Type[ForwardRef("C")]},
        {"~_TV3", ForwardRef("A")},
        {"~_TV4", ForwardRef("B")},
        {"~_TV5", ForwardRef("C")},
    ]

    # _TNodes should be unchanged after unification
    compare_list_sets(actual_set, expected_set)

    if draw:
        gen_graph_from_nodes(tc._nodes)


def test_userdefn_inheritance_multilevel(draw=False):
    src = """
    class A:
        pass

    class B(A):
        pass

    class C(B):
        pass

    a = A()
    b = B()
    c = C()
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tc = ti.type_constraints
    a, b, c = [
        ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)
    ]

    assert tc.unify(b, a).getValue() == ForwardRef("B")
    assert tc.unify(c, b).getValue() == ForwardRef("C")
    assert tc.unify(c, a).getValue() == ForwardRef("C")
    assert isinstance(ti.type_constraints.unify(b, c), TypeFail)

    if draw:
        gen_graph_from_nodes(tc._nodes)


def test_builtin_abst_inheritance(draw=False):
    src = """
    x = 3
    # float takes in an argument of type SupportsFloat
    y = float(x)
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    tc = ti.type_constraints
    actual_set = tc_to_disjoint(tc)
    assert {"~_TV0", int} in actual_set
    assert {"~_TV1", float} in actual_set
    if draw:
        gen_graph_from_nodes(tc._nodes)


# TODO: more tests for builtin inheritance


# Inheritance-based method lookup


def test_userdefn_mro_simple(draw=False):
    src = """
    class A:
        def foo(self):
            return 0

    class B(A):
        pass

    b = B()
    x = b.foo()
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    _, b, x = [
        ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)
    ]
    assert ti.type_constraints.resolve(x).getValue() == int
    if draw:
        gen_graph_from_nodes(ti.type_constraints._nodes)


def test_userdefn_mro_multilevel(draw=False):
    src = """
    class A:
        def foo(self):
            return 0

    class B:
        pass

    class C(A):
        pass

    class D(B, C):
        pass

    d = D()
    x = d.foo()
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    _, d, x = [
        ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)
    ]
    assert ti.type_constraints.resolve(x).getValue() == int
    if draw:
        gen_graph_from_nodes(ti.type_constraints._nodes)


def test_userdefn_mro_diamond(draw=False):
    src = """
    class A:
        pass

    class B(A):
        def foo(self):
            return 'a'

    class C(A):
        def foo(self):
            return 0

    class D(B,C):
        pass

    d = D()
    x = d.foo() # this is a call to B.foo()
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    _, _, d, x = [
        ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)
    ]
    assert ti.type_constraints.resolve(x).getValue() == str
    if draw:
        gen_graph_from_nodes(ti.type_constraints._nodes)


@pytest.mark.skip(reason="Cannot yet handle Generic Protocols (SupportsAbs[T])")
def test_builtin_abst_base_mro(draw=False):
    src = """
    x = 3 # x descends from SupportsAbs[int]
    y = abs(x)
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    x, y = [ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)]
    assert ti.type_constraints.resolve(y).getValue() == int
    if draw:
        gen_graph_from_nodes(ti.type_constraints._nodes)


def test_builtin_call_simple_mro(draw=False):
    src = """
    class A:
        pass

    a = A()
    x = repr(a)
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    a, x = [ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)]
    assert ti.type_constraints.resolve(x).getValue() == str
    if draw:
        gen_graph_from_nodes(ti.type_constraints._nodes)


def test_builtin_binop_inheritance(draw=False):
    src = """
    x = 1.0 + 3
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    x = [ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)][0]
    assert ti.type_constraints.resolve(x).getValue() == float
    if draw:
        gen_graph_from_nodes(ti.type_constraints._nodes)


def test_builtin_comp_inheritance(draw=False):
    src = """
    x = (3 == 'abc')
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    x = [ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)][0]
    assert ti.type_constraints.resolve(x).getValue() == bool
    if draw:
        gen_graph_from_nodes(ti.type_constraints._nodes)


def test_userdefn_inherits_from_builtin(draw=False):
    src = """
    class MyStr(str):
        pass

    s = MyStr()
    x = s.lower()
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    s, x = [ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)]
    assert ti.type_constraints.resolve(x).getValue() == str
    if draw:
        gen_graph_from_nodes(ti.type_constraints._nodes)


def test_userdefn_overrides_builtin(draw=False):
    src = """
    class MyStr(str):
        def lower(self):
            return 0

    s = MyStr()
    x = s.lower()
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    _, s, x = [
        ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)
    ]
    assert ti.type_constraints.resolve(x).getValue() == int
    if draw:
        gen_graph_from_nodes(ti.type_constraints._nodes)


def test_builtin_generic_inheritance_init(draw=False):
    src = """
    x = set([1,2,3])
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    x = [ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)][0]
    assert ti.type_constraints.resolve(x).getValue() == Set[int]


def test_builtin_generic_inheritance_method_lookup(draw=False):
    src = """
    x = set([1,2,3])
    y = x.symmetric_difference([2,3])
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    x, y = [ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)]
    assert ti.type_constraints.resolve(y).getValue() == Set[int]
    if draw:
        gen_graph_from_nodes(ti.type_constraints._nodes)


def test_builtin_generic_inheritance_overloaded_init(draw=False):
    src = """
    x = set([1,2,3])
    y = list(x)
    """
    ast_mod, ti = cs._parse_text(src, reset=True)
    x, y = [ti.lookup_typevar(node, node.name) for node in ast_mod.nodes_of_class(nodes.AssignName)]
    assert ti.type_constraints.resolve(y).getValue() == List[int]
    if draw:
        gen_graph_from_nodes(ti.type_constraints._nodes)
